/*
 *			    Get Interface
 *			    -------------
 *
 *		Jean II '00
 *
 * Get the interface of the other node.
 * The other node must run an e-Squirt app with the get interface handler set,
 * such as squirt.c...
 * This version use the official API...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/compat/CEGetInterfaceSample.c,v 2.7 2001/05/09 21:30:39 jeant Exp $
 * $Log: CEGetInterfaceSample.c,v $
 * Revision 2.7  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.6  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 * Revision 2.5  2001/04/21 02:22:21  jeant
 * RUI goodness (unescape/parse) and associated test programs. Updated official API from latest Palm header. More bugs fixed. Still not fully tested.
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <CESquirt.h>

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  TESquirtOption	how;
  TCoolFileInfo		info;
  TCoolBuffer *		data;
  char			name[256];
  char			type[256];
  int			error;

  /* Init */
  name[0] = '\0';
  type[0] = '\0';
  info.fileName = NULL;
  info.fileType = NULL;

  /* Look if there is any command line args */
  switch(argc)
    {
    case 0:
      /* Fallthrough */
    case 1:
      /* No argument : no name or type */
      break;
    case 2:
      /* One argument : the filename */
      strcpy(name, argv[1]);
      break;
    default:
      /* Fallthrough */
    case 3:
      /* Two arguments : filename + type */
      strcpy(name, argv[1]);
      strcpy(type, argv[2]);
      break;
    }

  /* Create buffers - should convert to unicode */
  if(name[0] != '\0')
    info.fileName = TCoolBuffer_Allocate((const unsigned char *) name,
					 strlen(name) + 1);
  if(type[0] != '\0')
    info.fileType = TCoolBuffer_Allocate((const unsigned char *) type,
					 strlen(type) + 1);

  /* Set the options */
  how.mode = ESQUIRT_AUTO;
  how.transport = ESQUIRT_AUTO;
  how.protocol = ESQUIRT_AUTO;

  /* Get the interface */
  data = CESquirt_GetControlPage(how, error, NULL, NULL, &info, NULL, NULL);

  /* Display it */
  if(data)
    {
      printf("Got [%s]\n", data->content);
      /* Won't work because of unicode... */
      if(info.fileName)
	printf("  -> name [%s]\n", info.fileName->content);
      if(info.fileType)
	printf("  -> type [%s]\n", info.fileType->content);
    }
  else
    printf("Get interface failed...\n");

  return(0);
}
