//
// CESquirt header file
//
// ESquirt-2 API
// PalmOS Implementation
// John Schettino (schettj@hpl.hp.com)
//

// CVS info
/*
$Header: /home/cvs/Hack/Esquirt/Linux/compat/CESquirt.h,v 2.6 2001/05/09 21:30:39 jeant Exp $
*/
/*
$Log: CESquirt.h,v $
Revision 2.6  2001/05/09 21:30:39  jeant
Proper licensing terms (thanks to Mirjana)...

Revision 2.5  2001/04/21 02:22:21  jeant
RUI goodness (unescape/parse) and associated test programs. Updated official API from latest Palm header. More bugs fixed. Still not fully tested.

Revision 2.4.3.1  2001/02/16 23:57:34  jeant
Imported sources

Revision 1.1.1.1  2001/02/15 00:09:16  johns
First version

*/
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#ifndef __CESQUIRT__
#define __CESQUIRT__

#include "TCoolReference.h"
#include "TESquirtOption.h"


// Authentication callback function interface
typedef bool (*AuthCallback)(char * &userName, char * &passWd);

// PalmOS Specific sendCallback
typedef void (*SendCallback)(TESquirtOption *how);

// PalmOS Specific recvCallback adds int errorCode
typedef void (*RecvCallback)(TESquirtOption *how, TCoolReference *data, int errorCode);

// PalmOS Specific controllCallback
typedef void (*ControllCallback)(TCoolBuffer *data);

// Standard send/receive status values
#define ESQUIRT_SEND_COMPLETE   0
#define ESQUIRT_SEND_PENDING    1
#define ESQUIRT_SEND_FAIL               -1

#define ESQUIRT_RECV_COMPLETE   0
#define ESQUIRT_RECV_PENDING    1
#define ESQUIRT_RECV_FAIL               -1

// PalmOS Specific receive status values
#define PALMOS_RECV_BUFFTOSMALL                 -2              // caller allocated TCoolRef buffer too small
#define PALMOS_RECV_PROVIDED_INTERFACE  -3              // Provided an interface during call
#define PALMOS_RECV_CANCELED                    -4              // RecvCancel() called

// File information structure
typedef struct _TCoolFileInfo {
        TCoolBuffer* fileName; // A file name, in Unicode
        TCoolBuffer* fileType; // A mime type (such as "text/html") in ascii
} TCoolFileInfo;


// Control Page request callback function interface
#define ESQUIRT_REQ_AUTH        1  // to request an authentication
#define ESQUIRT_OK                      0  // to indicate payload is valid and should be returned
#define ESQUIRT_FAIL            -1 // for either (a) fName not found or (b) auth failed

// Note: return value from this function is any valid Obex response code
// With the final bit set (ie, the number in ()s below)
/* From the OBEX spec:
0x10 (0x90) 100 Continue
0x20 (0xA0) 200 OK, Success
0x21 (0xA1) 201 Created
0x22 (0xA2) 202 Accepted
0x23 (0xA3) 203 Non-Authoritative Information
0x24 (0xA4) 204 No Content
0x25 (0xA5) 205 Reset Content
0x26 (0xA6) 206 Partial Content
0x30 (0xB0) 300 Multiple Choices
0x31 (0xB1) 301 Moved Permanently
0x32 (0xB2) 302 Moved temporarily
0x33 (0xB3) 303 See Other
0x34 (0xB4) 304 Not modified
0x35 (0xB5) 305 Use Proxy
0x40 (0xC0) 400 Bad Request - server couldn't understand request
0x41 (0xC1) 401 Unauthorized
0x42 (0xC2) 402 Payment required
0x43 (0xC3) 403 Forbidden - operation is understood but refused
0x44 (0xC4) 404 Not Found
0x45 (0xC5) 405 Method not allowed
0x46 (0xC6) 406 Not Acceptable
0x47 (0xC7) 407 Proxy Authentication required
0x48 (0xC8) 408 Request Time Out
0x49 (0xC9) 409 Conflict
0x4A (0xCA) 410 Gone
0x4B (0xCB) 411 Length Required
0x4C (0xCC) 412 Precondition failed
0x4D (0xCD) 413 Requested entity too large
0x4E (0xCE) 414 Request URL too large
0x4F (0xCF) 415 Unsupported media type
0x50 (0xD0) 500 Internal Server Error
0x51 (0xD1) 501 Not Implemented
0x52 (0xD2) 502 Bad Gateway
0x53 (0xD3) 503 Service Unavailable
0x54 (0xD4) 504 Gateway Timeout
0x55 (0xD5) 505 HTTP version not supported
*/
// return value of 0xC1 means begin auth. process (not implemented in this version)
// return value of 0xA0 means page valid
// any other return value means no control page returned
#define ESQUIRT_OBEX_OK                 0xa0  // Obex "success"   200 ok
#define ESQUIRT_OBEX_NO_CONTENT 0xa4    // 204 No content, but request processed ok (Get W/Ref)
#define ESQUIRT_OBEX_NOTFOUND   0xc4  // Obex "not found" 404 error

typedef int (*ControlPageCallback)(char * &userName, TCoolReference *data, TCoolFileInfo* info, TCoolBuffer * &pageContent);


// define functions to send/receive/get/provide
// Send
// PalmOS specific -- it is legal to "send" even when async Receive is pending
int  CESquirt_Send(TCoolReference &data, TESquirtOption &how, AuthCallback auth=0, SendCallback send=0);
int  CESquirt_SendStatus(); // returns status of request made above
void CESquirt_SendCancel(); // cancels request made above

// Receive
// PalmOS specific --  "wait" parm. Default is true (blocking call)
// If wait is false, call returns right away and receive happens in background (non-blocking)
// BUT in order for this to work, you must have a call to CESquirt_RecvHandleEvent() in your application event loop.
int  CESquirt_Recv(TCoolReference &data, TESquirtOption &how, RecvCallback recv=0, bool wait=true);
int  CESquirt_RecvStatus(); // returns status of request made above
void CESquirt_RecvCancel(int = PALMOS_RECV_CANCELED); // cancels request made above

// PalmOS Specific Functions
#ifndef __unix__
// PalmOS Specific calls to control 'Automatic Beam Recipt'
void DisablePalmBeam();         // disable receive beam if it is enabled
void RestorePalmBeam();         // restore receive beam

// Handle events generated by async CESquirt_Recv() call.
Boolean CESquirt_RecvHandleEvent(EventPtr eventP);

// parse a URL from EudoraWeb
int CESquirt_RUI(char * url, TCoolReference * &data, TCoolBuffer * &target, TCoolFileInfo * &info);
// save an html page into the memo pad, return the record number
unsigned short CESquirt_SavePage(TCoolBuffer* html);
// count (or purge) records from the notepad html page cache
unsigned int CESquirt_ManagePages(unsigned short& numCache, int delCache);
void CESquirt_LaunchEudoraWeb(unsigned short memooffset);
#endif


// Request a Control Page
TCoolBuffer *CESquirt_GetControlPage(TESquirtOption &how, int &error, TCoolReference *data=0, TCoolBuffer *target=0, TCoolFileInfo* info=0, AuthCallback auth=0, ControllCallback cont=0);              // request control page
int  CESquirt_GetStatus();              // returns status of request made above
void CESquirt_GetCancel();              // cancels request made above

// Provide a Control Page
// PalmOS: unicodeFileName ignored, so pageRequest() called for all requests
int CESquirt_ProvideControlPage(TESquirtOption &how, TCoolBuffer* unicodeFileName, ControlPageCallback pageRequest);    // provide control page
void CESquirt_ProvideCancel();                          // Stop providing control page
bool CESquirt_Authenticate(char *passwd);       // authenticate password for pending HandleControlPage()
// PalmOS: authentication not implemented in this release

// Get Capabilities
void  CESquirt_GetCapabilities(TESquirtOption &how);

#endif
