/*
 *			    Generic Squirt
 *			    --------------
 *
 *		Jean II '99 & '00
 *
 * Send and receive e-Squirt using all options...
 * This version use the official API...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/compat/CESquirtSample.c,v 2.7 2001/05/21 21:34:23 jeant Exp $
 * $Log: CESquirtSample.c,v $
 * Revision 2.7  2001/05/21 21:34:23  jeant
 * Finishing up RUI changes (obex_type, name) and finish update compat.
 *
 * Revision 2.6  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.5  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <CESquirt.h>

/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon.
 */
void receive_esquirt(TESquirtOption *how, TCoolReference *data, int errorCode)
{
  /* Display content */
  if(how->mode == ESQUIRT_BEACON)
    printf("Received a beacon : [%s]\n", data->content);
  else
    printf("Received a squirt : [%s]\n", data->content);

  /* Cleanup */
  TCoolReference_Free(data);
  free(how);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a Get Interface request.
 */
int reply_obget(char * &userName, TCoolReference *data, TCoolFileInfo* info,  TCoolBuffer * &pageContent)
{
  static char	string[] = "<HTML><HEAD><TITLE>No Interface</TITLE></HEAD>
<BODY>Please come back later.</BODY></HTML>";

  /* Print the squirt if present */
  if(data)
    printf("Received a reference : [%s]\n", data->content);

  /* Print name and type if there - should convert name from unicode */
  if((info != NULL) && (info->fileName != NULL))
    printf("fileName = [%s]\n", info->fileName->content);
  if((info != NULL) && (info->fileType != NULL))
    printf("fileType = [%s]\n", info->fileType->content);
  /* No way to get target here... */

  /* Print RUI data if present */
  if((pageContent != NULL) && (pageContent->content != NULL))
    printf("Received RUI data : [%s]\n", pageContent->content);

  /* Get rid of potential form data passed to us (only after processed it) */
  if(pageContent != NULL)
    TCoolBuffer_Free(pageContent);

  /* Malloc some data */
  pageContent = TCoolBuffer_Allocate((const unsigned char *) string,
				     strlen(string) + 1);

  /* Return the data back */
  return(0xA0);
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  TCoolReference *	data;
  TESquirtOption	how;
  int		ret = -1;

  /* Do allocation */
  data = TCoolReference_Allocate(NULL, 1024);

  /* Set the options */
  how.mode = ESQUIRT_AUTO;
  how.transport = ESQUIRT_AUTO;
  how.protocol = ESQUIRT_AUTO;

  /* Look if there is any command line args */
  switch(argc)
    {
    case 0:
      /* Fallthrough */
    case 1:
      /* No argument : receive a beacon */
      break;
    default:
      /* Fallthrough */
    case 2:
      /* One argument : the XML */
      strcpy((char *) data->content, argv[1]);
      break;
    }

  if(argc > 1)
    {
      /* Send a single e-Squirt/Beacon */
      how.mode = ESQUIRT_SQUIRT;	/* Can't have auto here ! */
      ret = CESquirt_Send(*data, how, NULL, NULL);
      if(ret < 0)
	printf("Sending e-Squirt failed...\n");
    }
  else
    {
      /* Prepare to handle get interface calls */
      CESquirt_ProvideControlPage(how, NULL, &reply_obget);

      /* Receive multiple e-Squirt/Beacon */
      printf("Waiting for e-Squirts over all transports...\n");
      CESquirt_Recv(*data, how, &receive_esquirt);
      /* Will loop forever */
    }

  return(ret);
}
