/*
 *			    Squirt main common
 *			    ------------------
 *
 *		Jean II '00
 *
 * Handle all the common code for the various e-squirt applications.
 * Main calls to handle e-Squirt...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/compat/CESquirtWrapper.c,v 2.14 2001/05/21 21:34:23 jeant Exp $
 * $Log: CESquirtWrapper.c,v $
 * Revision 2.14  2001/05/21 21:34:23  jeant
 * Finishing up RUI changes (obex_type, name) and finish update compat.
 *
 * Revision 2.13  2001/05/18 02:43:57  jeant
 * First pass on latest GET/FORM encoding and overflows. Still need more testing and a few questions answered...
 *
 * Revision 2.12  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.11  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 * Revision 2.10  2001/04/21 02:22:21  jeant
 * RUI goodness (unescape/parse) and associated test programs. Updated official API from latest Palm header. More bugs fixed. Still not fully tested.
 *
 * Revision 2.9  2001/03/06 23:47:06  jeant
 * Version 2.4.4 (don't trust CVS) : Properly terminate XML ('\>' -> '/>'), Allow sending no-session GET, mandate OpenObex 0.9.8 and minor cleanups.
 *
 * Revision 2.8  2001/02/21 21:54:38  jeant
 * Oups ! Minor bug fix : control -> &control
 *
 * Revision 2.7  2001/02/21 21:48:46  jeant
 * Add select handler in compat/ to ignore signals
 *
 * Revision 2.6  2001/02/21 03:04:52  jeant
 * Checking what's happening with version numbers...
 *
 * Revision 2.5  2001/02/21 02:58:59  jeant
 * Forgot to make RecvCancel comformat to new API :-(
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include <sys/socket.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <linux/types.h>
#include <linux/irda.h>

#include "squirt.h"

#include "CESquirt.h"

/**************************** VARIABLES ****************************/
/*
 * Try to be light there...
 */
/* Control page callback */
static ControlPageCallback recv_pageRequest = NULL;

/************************ TCOOLBUFFER CALLS ************************/
/*
 * Implementation of all the routines defined in TCoolReference.h
 */

/* ---------------------------------------------------------------- */
/*
 * malloc an empty ref, no content memory
 */
TCoolBuffer * TCoolBuffer_Allocate()
{
  TCoolBuffer *	ref;

  /* Just do it ! */
  ref = (TCoolBuffer *) malloc(sizeof(TCoolBuffer));
  if(ref)
    {
      ref->content = NULL;
      ref->length = 0;
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * malloc a ref & ref->content, copy in content
 */
TCoolBuffer * TCoolBuffer_Allocate(const TCoolBuffer& ref)
{
  TCoolBuffer *	ref2;

  /* Just do it ! */
  ref2 = (TCoolBuffer *) malloc(sizeof(TCoolBuffer));
  if(ref2)
    {
      ref2->content = (unsigned char *) malloc(ref.length);
      memcpy(ref2->content, ref.content, ref.length);
      ref2->length = ref.length;
    }

  return(ref2);
}

/* ---------------------------------------------------------------- */
/*
 * malloc ref & ref->content, copy in val
 */
TCoolBuffer * TCoolBuffer_Allocate(const unsigned char *value, unsigned int len)
{
  TCoolBuffer *	ref;

  /* Just do it ! */
  ref = (TCoolBuffer *) malloc(sizeof(TCoolBuffer));
  if(ref)
    {
      ref->content = (unsigned char *) malloc(len);
      memset(ref->content, '\0', len);
      if(value != NULL)
	memcpy(ref->content, value, len);
      ref->length = len;
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * malloc ref & ref->content, unintialized
 */
TCoolBuffer * TCoolBuffer_Allocate(unsigned int len)
{
  TCoolBuffer *	ref;

  /* Just do it ! */
  ref = (TCoolBuffer *) malloc(sizeof(TCoolBuffer));
  if(ref)
    {
      ref->content = (unsigned char *) malloc(len);
      memset(ref->content, '\0', len);
      ref->length = len;
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * free ref->content and then ref
 */
void TCoolBuffer_Free(TCoolBuffer *ref)
{
  if(ref)
    {
      /* Free content */
      if(ref->content)
	free(ref->content);
      ref->content = NULL;
      /* Free reference */
      free(ref);
    }
}

/*********************** TCOOLREFERENCE CALLS ***********************/
/*
 * Implementation of all the routines defined in TCoolReference.h
 */

/* ---------------------------------------------------------------- */
/*
 * malloc an empty ref, no content memory
 */
TCoolReference * TCoolReference_Allocate()
{
  TCoolReference *	ref;

  /* Just do it ! */
  ref = (TCoolReference *) malloc(sizeof(TCoolReference));
  if(ref)
    {
      ref->content = NULL;
      ref->length = 0;
      ref->encoding = NULL;
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * malloc a ref & ref->content, copy in content
 */
TCoolReference * TCoolReference_Allocate(const TCoolReference& ref)
{
  TCoolReference *	ref2;

  /* Just do it ! */
  ref2 = (TCoolReference *) malloc(sizeof(TCoolReference));
  if(ref2)
    {
      ref2->content = (unsigned char *) malloc(ref.length);
      memcpy(ref2->content, ref.content, ref.length);
      ref2->length = ref.length;
      ref2->encoding = NULL;
    }

  return(ref2);
}

/* ---------------------------------------------------------------- */
/*
 * malloc ref & ref->content, copy in val
 */
TCoolReference * TCoolReference_Allocate(const unsigned char *value, unsigned int len, char * encode)
{
  TCoolReference *	ref;

  /* Just do it ! */
  ref = (TCoolReference *) malloc(sizeof(TCoolReference));
  if(ref)
    {
      ref->content = (unsigned char *) malloc(len);
      memset(ref->content, '\0', len);
      if(value != NULL)
	memcpy(ref->content, value, len);
      ref->length = len;

      ref->encoding = NULL;
      if(encode != NULL)
	{
	  ref->encoding = (char *) malloc(strlen(encode) + 1);
	  memcpy(ref->encoding, encode, strlen(encode) + 1);
	}
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * malloc ref & ref->content, unintialized
 */
TCoolReference * TCoolReference_Allocate(unsigned int len)
{
  TCoolReference *	ref;

  /* Just do it ! */
  ref = (TCoolReference *) malloc(sizeof(TCoolReference));
  if(ref)
    {
      ref->content = (unsigned char *) malloc(len);
      memset(ref->content, '\0', len);
      ref->length = len;
      ref->encoding = NULL;
    }

  return(ref);
}

/* ---------------------------------------------------------------- */
/*
 * free ref->content and then ref
 */
void TCoolReference_Free(TCoolReference *ref)
{
  if(ref)
    {
      /* Free content */
      if(ref->content)
	free(ref->content);
      ref->content = NULL;
      /* Free encoding */
      if(ref->encoding)
	free(ref->encoding);
      ref->content = NULL;
      /* Free reference */
      free(ref);
    }
}


/************************* DATA TRANSLATION*************************/
/*
 * Translate between the native and the API representation of the data.
 */

/* ---------------------------------------------------------------- */
/*
 * Translate a (TCoolReference *) to a (es_squirt *) and translate a
 * (TESquirtOption *) to a (es_opt *)
 */
static inline int
TCool_to_squirt(TCoolReference *	tref,
		TESquirtOption *	topt,
		es_squirt *		content,	/* Payload */
		es_opt *		options,	/* How to process */
		es_trans *		extra)		/* Transport data */
{
  /* Cleanup */
  init_structs(content, options, extra, NULL, NULL, NULL);

  /* The TESquirtOption */
  if((topt != NULL) && (options != NULL))
    {
      /* Defaults to API compliant - do not change that */
      options->do_xml = OPT_NO_XML;
      options->obex.ias = OPT_OBEX_NO_STDIAS;
      //options->obex.name = OPT_OBEX_UNICODE_NAME;	/* Recv */
      options->obex.name = OPT_OBEX_NO_NAME;		/* Send */
      options->obex.type = OPT_OBEX_TYPE_REFX;
      options->obex.session = OPT_OBEX_DO_SESSION;
      options->obex.target = OPT_OBEX_TARGET_UUGLI;
      options->get_iface = OPT_NO_GIFACE;
      options->uobex_dups = OPT_UOS_DUPS_DEL;

      /* The model */
      switch(topt->mode)
	{
	case ESQUIRT_SQUIRT:
	  options->model = OPT_SQUIRT;
	  break;
	case ESQUIRT_BEACON:
	  options->model = OPT_BEACON;
	  break;
	default:
	  options->model = OPT_UNDEFINED;
	}

      /* The transport */
      switch(topt->transport)
	{
	case ESQUIRT_ULTRA:
	  options->transport = OPT_ULTRA;
	  break;
	case ESQUIRT_IRDA:
	  options->transport = OPT_IRDA;
	  break;
	default:
	  options->transport = OPT_UNDEFINED;
	}

      /* The protocol */
      switch(topt->protocol)
	{
	case ESQUIRT_RAW:
	  options->protocol = OPT_RAW;
	  break;
	case ESQUIRT_OBEX:
	  options->protocol = OPT_OBEX;
	  break;
	case ESQUIRT_OBEX_GET:
	  options->protocol = OPT_OBEX_GET;
	  break;
	default:
	  options->protocol = OPT_UNDEFINED;
	}
    }

  /* The TCoolReference content */
  if((tref != NULL) && (content != NULL))
    {
      /* Check overflow */
      if(tref->length > ES_MAX_XML)
	return(-1);
      /* Copy over */
      memcpy(content->xml, tref->content, tref->length);
      /* NULL Terminate the string */
      if(tref->length != ES_MAX_XML)
	content->xml[tref->length] = '\0';
      else
	content->xml[ES_MAX_XML - 1] = '\0';
    }

  /* Now the TCoolReference encoding */
  if((tref != NULL) && (tref->encoding != NULL) && (extra != NULL))
    {
      /* Check overflow */
      if((strlen(tref->encoding) + 1) > ES_MAX_TYPE)
	return(-1);
      /* Copy over */
      memcpy(extra->type, tref->encoding, strlen(tref->encoding) + 1);

      /* Options */
      options->obex.type = OPT_OBEX_TYPE_USER;
    }
      
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Reverse of the previous function.
 * Translate a (es_squirt *) to a (TCoolReference *) and translate a
 * (es_opt *) to a (TESquirtOption *)
 */
static inline int
squirt_to_TCool(es_squirt *		content,	/* Payload */
		es_opt *		options,	/* How to process */
		es_trans *		extra,		/* Transport data */
		TCoolReference *	tref,
		TESquirtOption *	topt)
{
  /* The TCoolReference content */
  if((tref != NULL) && (content != NULL))
    {
      int	max_len = ES_MAX_XML;

      /* Check overflow */
      if(tref->length < ES_MAX_XML)
	max_len = tref->length;
      /* Copy over (string is already NULL terminated) */
      memcpy(tref->content, content->xml, max_len);
      tref->content[max_len - 1] = '\0';
    }

  /* The TCoolReference encoding */
  if((tref != NULL) && (extra != NULL))
    {
      /* If extra is valid, all type will also be avail textually */
      if(options->obex.type >= OPT_OBEX_TYPE_USER)
	{
	  tref->encoding = (char *) malloc(strlen(extra->type) + 1);
	  memcpy(tref->encoding, extra->type, strlen(extra->type) + 1);
	}
    }

  /* The TESquirtOption */
  if((topt != NULL) && (options != NULL))
    {
      /* The model */
      switch(options->model)
	{
	case OPT_SQUIRT:
	  topt->mode = ESQUIRT_SQUIRT;
	  break;
	case OPT_BEACON:
	  topt->mode = ESQUIRT_BEACON;
	  break;
	default:
	  topt->mode = ESQUIRT_AUTO;
	}

      /* The transport */
      switch(options->transport)
	{
	case OPT_ULTRA:
	  topt->transport = ESQUIRT_ULTRA;
	  break;
	case OPT_IRDA:
	  topt->transport = ESQUIRT_IRDA;
	  break;
	default:
	  topt->transport = ESQUIRT_AUTO;
	}

      /* The protocol */
      switch(options->protocol)
	{
	case OPT_RAW:
	  topt->protocol = ESQUIRT_RAW;
	  break;
	case OPT_OBEX:
	  topt->protocol = ESQUIRT_OBEX;
	  break;
	case OPT_OBEX_GET:
	  topt->protocol = ESQUIRT_OBEX_GET;
	  break;
	default:
	  topt->protocol = ESQUIRT_AUTO;
	}
    }

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Translate a (TCoolFileInfo *) and a (TCoolReference *) to a (gi_obget *)
 * and translate a (TESquirtOption *) to a (gi_flags *)
 */
static inline int
TCool_to_obget(TCoolFileInfo *		tinfo,
	       TCoolBuffer *		tdata,
	       TCoolBuffer *		ttarget,
	       TESquirtOption *		topt,
	       gi_obget *		request,
	       gi_flags *		flags,
	       es_trans *		extra,
	       int			reply)
{
  /* Cleanup */
  init_structs(NULL, NULL, extra, request, flags, NULL);

  /* The TESquirtOption */
  if((topt != NULL) && (flags != NULL))
    {
      /* Defaults to API compliant - do not change that */
      flags->obex.ias = OPT_OBEX_NO_STDIAS;
      flags->obex.name = OPT_OBEX_NO_NAME;
      flags->obex.type = OPT_OBEX_TYPE_USER;
      flags->obex.session = OPT_OBEX_DO_SESSION;
      /* Note : some will be overriden just below */

      /* The transport */
      switch(topt->transport)
	{
	case ESQUIRT_ULTRA:
	  /* Not available */
	  flags->transport = OPT_ULTRA;
	  break;
	case ESQUIRT_IRDA:
	  flags->transport = OPT_IRDA;
	  break;
	default:
	  flags->transport = OPT_UNDEFINED;
	}

      /* The protocol */
      switch(topt->protocol)
	{
	case ESQUIRT_RAW:
	  /* Not available */
	  flags->protocol = OPT_RAW;
	  break;
	case ESQUIRT_OBEX:
	  flags->protocol = OPT_OBEX;
	  break;
	case ESQUIRT_OBEX_GET:
	  flags->protocol = OPT_OBEX_GET;
	  break;
	default:
	  flags->protocol = OPT_UNDEFINED;
	}
    }

  /* The TCoolBuffer *ttarget */
  if((ttarget != NULL) && (extra != NULL))
    {
      /* Check overflow */
      if(ttarget->length > ES_MAX_TARGET)
	return(-1);
      /* Copy over */
      memcpy(extra->target, ttarget->content,
	     ttarget->length);
      /* Set the length */
      extra->targetlen = ttarget->length;
      /* Set the flag proper */
      flags->obex.target = OPT_OBEX_TARGET_USER;
    }

  /* The TCoolFileInfo->fileName */
  if((tinfo != NULL) && (tinfo->fileName != NULL) && (extra != NULL))
    {
      /* Check overflow */
      if(tinfo->fileName->length > ES_MAX_FILENAME)
	return(-1);
      /* Copy over */
      memcpy(extra->filename, tinfo->fileName->content,
	     tinfo->fileName->length);
      /* NULL Terminate the string */
      if(tinfo->fileName->length != ES_MAX_FILENAME)
	extra->filename[tinfo->fileName->length] = '\0';
      else
	extra->filename[ES_MAX_FILENAME - 1] = '\0';
      /* Set the flag proper */
      flags->obex.name = OPT_OBEX_UNICODE_NAME;
    }

  /* The TCoolFileInfo->filetype */
  if((tinfo != NULL) && (tinfo->fileType != NULL) && (extra != NULL))
    {
      /* If it is a request, must go in HTTP-Accept field,
       * If it is a reply, must go in Type field */
      if(reply)
	{
	  /* Check overflow */
	  if(tinfo->fileType->length > ES_MAX_TYPE)
	    return(-1);
	  /* Copy over */
	  memcpy(extra->type, tinfo->fileType->content,
		 tinfo->fileType->length);
	  /* NULL Terminate the string */
	  if(tinfo->fileType->length != ES_MAX_TYPE)
	    extra->type[tinfo->fileType->length] = '\0';
	  else
	    extra->type[ES_MAX_TYPE - 1] = '\0';
	  /* Set the flag proper */
	  flags->obex.type = OPT_OBEX_TYPE_USER;
	}
      else
	{
	  /* Check overflow */
	  if(tinfo->fileType->length > ES_MAX_ACCEPT)
	    return(-1);
	  /* Copy over */
	  memcpy(extra->accept, tinfo->fileType->content,
		 tinfo->fileType->length);
	  /* NULL Terminate the string */
	  if(tinfo->fileType->length != ES_MAX_ACCEPT)
	    extra->accept[tinfo->fileType->length] = '\0';
	  else
	    extra->accept[ES_MAX_ACCEPT - 1] = '\0';
	  /* Set the flag proper */
	  flags->obex.accept = OPT_OBEX_HAS_ACCEPT;
	}
    }

  /* The TCoolReference */
  if((tdata != NULL) && (request != NULL))
    {
      /* Easy : just point to it !!! - Warning on deallocation */
      request->data = (char *) tdata->content;
      request->datalen = tdata->length;
      request->cleanup = OPT_DATA_DO_FREE;
    }

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Reverse of the previous function.
 * Translate a (gi_obget *) to a (TCoolFileInfo *) and a (TCoolReference *)
 * and translate a (gi_flags *) to a (TESquirtOption *)
 */
static inline int
obget_to_TCool(gi_obget *		request,
	       gi_flags *		flags,
	       es_trans *		extra,
	       TCoolFileInfo *		tinfo,
	       TCoolBuffer *		tdata,
	       TCoolBuffer *		ttarget,
	       TESquirtOption *		topt,
	       int			reply)
{
  /* The TCoolBuffer *ttarget */
  if((ttarget != NULL) && (extra != NULL) &&
     (flags->obex.target == OPT_OBEX_TARGET_USER))
    {
      /* Check if the buffer is big enough */
      if(ttarget->length < (unsigned) extra->targetlen)
	{
	  free(ttarget->content);
	  ttarget->content = NULL;
	}

      /* Check if the buffer exist */
      if(ttarget->content == NULL)
	{
	  ttarget->content = (unsigned char *) malloc(extra->targetlen);
	  if(ttarget == NULL)
	    return(-1);
	  ttarget->length = extra->targetlen;
	}

      /* Copy over */
      memcpy(ttarget->content, extra->target, extra->targetlen);
    }

  /* The TCoolFileInfo->fileName */
  if((tinfo != NULL) && (extra != NULL) &&
     (flags->obex.name == OPT_OBEX_UNICODE_NAME))
    {
      /* Check if the structure is big enough */
      if((tinfo->fileName != NULL) &&
	 (tinfo->fileName->length < ES_MAX_FILENAME))
	{
	  TCoolBuffer_Free(tinfo->fileName);
	  tinfo->fileName = NULL;
	}

      /* Check if the structure exist */
      if(tinfo->fileName == NULL)
	{
	  tinfo->fileName = TCoolBuffer_Allocate(ES_MAX_FILENAME);
	  if(tinfo->fileName == NULL)
	    return(-1);
	}

      /* Copy over (string is already NULL terminated) */
      memcpy(tinfo->fileName->content, extra->filename, ES_MAX_FILENAME);
    }

  /* If it is a request, must fetch type from HTTP-Accept field,
   * If it is a reply, must fetch type from Type field */
  if(reply)
    {
      /* The TCoolFileInfo->fileType */
      if((tinfo != NULL) && (extra != NULL) &&
	 (flags->obex.type >= OPT_OBEX_TYPE_USER))
	{
	  /* Check if the structure is big enough */
	  if((tinfo->fileType != NULL) &&
	     (tinfo->fileType->length < ES_MAX_TYPE))
	    {
	      TCoolBuffer_Free(tinfo->fileType);
	      tinfo->fileType = NULL;
	    }

	  /* Check if the structure exist */
	  if(tinfo->fileType == NULL)
	    {
	      tinfo->fileType = TCoolBuffer_Allocate(ES_MAX_TYPE);
	      if(tinfo->fileType == NULL)
		return(-1);
	    }

	  /* Copy over (string is already NULL terminated) */
	  memcpy(tinfo->fileType->content, extra->type, ES_MAX_TYPE);
	}
    }
  else
    {
      /* The TCoolFileInfo->fileType */
      if((tinfo != NULL) && (extra != NULL) &&
	 (flags->obex.accept == OPT_OBEX_HAS_ACCEPT))
	{
	  /* Check if the structure is big enough */
	  if((tinfo->fileType != NULL) &&
	     (tinfo->fileType->length < ES_MAX_ACCEPT))
	    {
	      TCoolBuffer_Free(tinfo->fileType);
	      tinfo->fileType = NULL;
	    }

	  /* Check if the structure exist */
	  if(tinfo->fileType == NULL)
	    {
	      tinfo->fileType = TCoolBuffer_Allocate(ES_MAX_ACCEPT);
	      if(tinfo->fileType == NULL)
		return(-1);
	    }

	  /* Copy over (string is already NULL terminated) */
	  memcpy(tinfo->fileType->content, extra->accept, ES_MAX_ACCEPT);
	}
    }

  /* The TCoolReference */
  if((tdata != NULL) && (request != NULL))
    {
      /* Easy : just point to it !!! - Warning on deallocation */
      tdata->content = (unsigned char *) request->data;
      tdata->length = request->datalen;
      /* Currently, the lib give us always OPT_DATA_DO_FREE... */
    }

  /* The TESquirtOption */
  if((topt != NULL) && (flags != NULL))
    {
      /* Inits */
      topt->mode = ESQUIRT_SQUIRT;

      /* The transport */
      switch(flags->transport)
	{
	case OPT_ULTRA:
	  topt->transport = ESQUIRT_ULTRA;
	  break;
	case OPT_IRDA:
	  topt->transport = ESQUIRT_IRDA;
	  break;
	default:
	  topt->transport = ESQUIRT_AUTO;
	}

      /* The protocol */
      switch(flags->protocol)
	{
	case OPT_RAW:
	  topt->protocol = ESQUIRT_RAW;
	  break;
	case OPT_OBEX:
	  topt->protocol = ESQUIRT_OBEX;
	  break;
	case OPT_OBEX_GET:
	  topt->protocol = ESQUIRT_OBEX_GET;
	  break;
	default:
	  topt->protocol = ESQUIRT_AUTO;
	}
    }

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Translate the sending return code
 */
static inline int
mangle_send_ret(int	ret)
{
  if(ret >= 0)
    return(ESQUIRT_SEND_COMPLETE);
  else
    {
      if(ret < -10)
	return(-ret);
      else
	return(ESQUIRT_SEND_FAIL);
    }
}

/* ---------------------------------------------------------------- */
/*
 * Translate the sending return code
 */
static inline int
mangle_recv_ret(int	ret)
{
  if(ret >= 0)
    return(ESQUIRT_RECV_COMPLETE);
  else
    {
      if(ret < -10)
	return(-ret);
      else
	return(ESQUIRT_RECV_FAIL);
    }
}


/**************************** CALLBACKS ****************************/
/*
 * Callback for the native esquirt library.
 * We just call the proper user callback...
 */

/* ---------------------------------------------------------------- */
/*
 * Call the get interface callback of our choice...
 * The (TCoolFileInfo *) belong to us, the user should not deallocate it.
 * We will free the (TCoolBuffer *) returned to us.
 * Note : The API mixes in random fashion references and pointers,
 *	which is not cool. Anyway...
 */
static int
CallProvide(TESquirtOption *	how,
	    TCoolReference *	ref,
	    gi_obget *		request,
	    gi_flags *		flags,
	    es_trans *		extra,
	    ControlPageCallback	callback)
{
  TCoolFileInfo		info;
  TCoolBuffer *		data;
  TCoolBuffer *		target;	/* Oups - no way to pass it to the user !!! */
  char *	username = NULL;
  int		ret = 0;

  /* Check if we have all that we need */
  if(callback != NULL)
    {
      /* Prepare TCool data structs */
      info.fileName = NULL;
      info.fileType = NULL;
      data = NULL;		/* Callback must allocate it */
      target = TCoolBuffer_Allocate();
      data = TCoolBuffer_Allocate();

      /* Translate headers */
      obget_to_TCool(request, flags, extra, &info, data, target, NULL, 0);

      /* We can't pass the target to the user, so scrap it ! */
      TCoolBuffer_Free(target);

      /* Call the callback */
      ret = callback(username, ref, &info, data);

      /* Do we have something to send back ? */
      if(((ret == ESQUIRT_OK) || (ret == ESQUIRT_OBEX_OK)) && (data != NULL))
	{
	  /* Translate reply */
	  TCool_to_obget(&info, data, NULL, NULL, request, flags, extra, 1);

	  /* Cleanup the TCool mess */
	  if(info.fileName != NULL)
	    TCoolBuffer_Free(info.fileName);
	  if(info.fileType != NULL)
	    TCoolBuffer_Free(info.fileType);
	  free(data);	/* Don't use TCoolBuffer_Free() because of content */

	  /* Return back to the server - do free() on content */
	  return(2);
	}
      else
	{
	  /* Error code... */
	  if(ret > 10)
	    return(-ret);
	}
   }

  /* Everything fine, but no data */
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon.
 */
int
receive_esquirt(es_squirt *	content,	/* Rx payload */
		es_opt *	options,	/* Rx options */
		es_trans *	extra,		/* Rx transport data */
		gi_obget *	ret_hdr,	/* Return filename */
		gi_flags *	ret_flags)	/* Return flags */
{
  RecvCallback		callback = (RecvCallback) extra->userdata;
  TCoolReference *	ref;
  TESquirtOption *	how;
  int		ret = 0;	/* Everything was ok */

  /* Do allocation */
  ref = TCoolReference_Allocate(ES_MAX_XML);
  how = (TESquirtOption *) malloc(sizeof(TESquirtOption));

  /* Check if we have all that we need */
  if((callback != NULL) && (ref != NULL) && (how != NULL))
    {
      /* Convert data and options */
      squirt_to_TCool(content, options, extra, ref, how);

      /* Check if it's an Obex-GET */
      if(options->protocol != OPT_OBEX_GET)
	{
	  /* Simple squirt - Call the callback */
	  callback(how, ref, ESQUIRT_RECV_COMPLETE);
	}
      else
	{
	  /* OBEX-GET -> Ask the user for a control page */
	  ret = CallProvide(how, ref, ret_hdr, ret_flags, extra,
			    recv_pageRequest);

	  /* Note : depending on which spec you read, this may be wrong... */
	  /* And we also have a squirt to deliver */
	  callback(how, ref, PALMOS_RECV_PROVIDED_INTERFACE);
	}
    }
  else
    {
      /* Cleanup */
      if(ref)
	TCoolReference_Free(ref);
      if(how)
	free(how);
    }

  /* Everything was ok */
  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a Get Interface request.
 */
int
reply_obget(gi_obget *	request,
	    gi_flags *	flags,
	    es_trans *	extra,
	    void *	userdata)
{
  /* Ask for the job to be done ;-) */
  return(CallProvide(NULL, NULL, request, flags, extra,
		    (ControlPageCallback) userdata));
}

/* ---------------------------------------------------------------- */
/*
 * Handler for timeout or signals in the select
 *
 * The API doesn't offer anything to map this function to, so we
 * just ignore those conditions and continue...
 */
int
select_esquirt(int		code,
	       void *		cbdata)		/* CallBack data */
{
  /* Values of code : 0=timeout, -1=signal */

  /* Here we could check if we need to exit the receiver main loop */
  return(0);		/* 0 = Continue ; 1 = Exit receiver loop */
}


/************************** CESQUIRT CALLS **************************/
/*
 * Implementation of all the routines defined in CESquirt.h
 */

/* ---------------------------------------------------------------- */
/*
 * Send
 * Note : the call is always blocking
 * Note : the callbacks are not used
 */
int  CESquirt_Send(TCoolReference &data, TESquirtOption &how, AuthCallback auth, SendCallback send)
{
  es_squirt	content;
  es_opt	options;
  es_trans	extra;
  int		ret;

  /* Translate data */
  ret = TCool_to_squirt(&data, &how, &content, &options, &extra);
  if(ret < 0)
    return(ESQUIRT_SEND_FAIL);


  /* Send a single e-Squirt/Beacon */
  ret = send_squirt(&content, &options, &extra, NULL, NULL);
  return(mangle_send_ret(ret));
}

/* ---------------------------------------------------------------- */
/*
 * returns status of request made above
 * Note : Not available
 */
int  CESquirt_SendStatus()
{
  return(ESQUIRT_SEND_COMPLETE);
}

/* ---------------------------------------------------------------- */
/*
 * cancels request made above
 * Note : Not available
 */
void CESquirt_SendCancel()
{
  //return(ESQUIRT_SEND_FAIL);
}

/* ---------------------------------------------------------------- */
/*
 * Send
 * Note : the call is always blocking
 * Note : if recv is !NULL, loop forever and call recv for each receive.
 * Note : if recv is NULL, do a single receive and returns the result.
 */
int  CESquirt_Recv(TCoolReference &data, TESquirtOption &how, RecvCallback recv, bool wait)
{
  es_squirt	content;
  es_opt	options;
  es_trans	extra;
  es_server_data	control;
  int		ret;

  /* We support only blocking */
  if(!wait)
    return(ESQUIRT_SEND_FAIL);

  /* Translate data */
  ret = TCool_to_squirt(&data, &how, &content, &options, &extra);
  if(ret < 0)
    return(ESQUIRT_SEND_FAIL);
  /* Fix for receiver */
  options.obex.name = OPT_OBEX_UNICODE_NAME;

  /* Is there any callback ? */
  if(recv != NULL)
    {
      /* Set up callback stuff */
      extra.userdata = (void *) recv;

      /* Set up a control structure to handle properly signals */
      memset(&control, 0, sizeof(es_server_data));
      control.timeback = &select_esquirt;
      /* No timeout, no accept callback, no userdata */

      /* Receive multiple e-Squirt/Beacon */
      recv_squirt(&content, &options, &extra, &receive_esquirt, &control);
      /* Will loop forever */
      ret = -1;
    }
  else
    {
      /* Receive a single e-Squirt/Beacon */
      ret = recv_squirt(&content, &options, NULL, NULL, NULL);

      /* Convert data and options */
      squirt_to_TCool(&content, &options, &extra, &data, &how);
    }

  return(mangle_recv_ret(ret));
}

/* ---------------------------------------------------------------- */
/*
 * returns status of request made above
 * Note : Not available
 */
int  CESquirt_RecvStatus()
{
  return(ESQUIRT_RECV_PENDING);
}

/* ---------------------------------------------------------------- */
/*
 * cancels request made above
 * Note : Not available
 */
void CESquirt_RecvCancel(int err)
{
  //return(ESQUIRT_RECV_FAIL);
}

/* ---------------------------------------------------------------- */
/*
 * Request a Control Page
 * Note : the call is always blocking
 * Note : The API mixes references and pointers in random fashion, and is
 *	not even consistent with CESquirt_Send(). Can you make it more
 *	confusing ?
 * Note : User is supposed to cleanup the returned TCoolBuffer and the
 *	eventual info->fileName and info->fileType created, as well as
 *	their TCoolBuffer->content.
 */
TCoolBuffer *CESquirt_GetControlPage(TESquirtOption &how, int &error,  TCoolReference *data, TCoolBuffer *target, TCoolFileInfo* info, AuthCallback auth, ControllCallback cont)
{
  gi_obget	interface;
  gi_flags	flags;
  es_trans	extra;
  TCoolBuffer *	page;
  int		ret;

  /* Translate data */
  ret = TCool_to_obget(info, NULL, target, &how,
		       &interface, &flags, &extra, 0);
  if(ret < 0)
    return(NULL);

  /* Check if there is a squirt to send */
  if(data == NULL)
    {
      /* Get the interface */
      ret = get_interface(&interface, &flags, &extra);
    }
  else
    {
      es_squirt	content;
      es_opt	options;

      /* Translate data */
      ret = TCool_to_squirt(data, &how, &content, &options, &extra);
      if(ret < 0)
	return(NULL);
      /* Make sure we send via a OBEX-GET */
      options.protocol = OPT_OBEX_GET;
      /* Fix/Hack for the receiver part */
      options.obex.name = OPT_OBEX_UNICODE_NAME;

      /* Send a single e-Squirt/Beacon and get interface */
      ret = send_squirt(&content, &options, &extra, &interface, &flags);
    }

  /* Set error code */
  error = mangle_send_ret(ret);

  /* Failure */
  if(ret < 0)
    return(NULL);

  /* Create a empty/unlinked buffer */
  page = TCoolBuffer_Allocate();
  if(page == NULL)
    return(NULL);

  /* Decapsulate the data on the way back */
  obget_to_TCool(&interface, &flags, &extra, info, page, NULL, &how, 1);
  return(page);
}

/* ---------------------------------------------------------------- */
/*
 * returns status of request made above
 * Note : Not available
 */
int  CESquirt_GetStatus()
{
  return(ESQUIRT_RECV_PENDING);
}

/* ---------------------------------------------------------------- */
/*
 * cancels request made above
 * Note : Not available
 */
void CESquirt_GetCancel()
{
  //return(ESQUIRT_RECV_FAIL);
}

/* ---------------------------------------------------------------- */
/*
 * Provide a Control Page
 * Note : the call is always non-blocking
 * Note : the server is active only is a receiver is active
 * Note : The options are not mapped to anything, because GET-Interface
 *	is fully dependant on the receiver options
 * Note : we don't support unicodeFileName
 */
int CESquirt_ProvideControlPage(TESquirtOption &how, TCoolBuffer* unicodeFileName, ControlPageCallback pageRequest)
{
  if(unicodeFileName != NULL)
    return(ESQUIRT_RECV_FAIL);

  if(pageRequest != NULL)
    {
      /* Prepare to handle get interface calls */
      set_obget_handler(&reply_obget, (void *) pageRequest);
      recv_pageRequest = pageRequest;
    }
  else
    {
      /* Do not handle get interface calls anymore */
      set_obget_handler(NULL, NULL);
      recv_pageRequest = NULL;
    }

  return(ESQUIRT_RECV_PENDING);
}

/* ---------------------------------------------------------------- */
/*
 * Stop providing control page
 */
void CESquirt_ProvideCancel()
{
  /* Do not handle get interface calls anymore */
  set_obget_handler(NULL, NULL);
  recv_pageRequest = NULL;

  //return(ESQUIRT_RECV_COMPLETE);
}

/* ---------------------------------------------------------------- */
/*
 * authenticate password for pending HandleControlPage()
 */
bool CESquirt_Authenticate(char *passwd)
{
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Get Capabilities
 */
void  CESquirt_GetCapabilities(TESquirtOption &how)
{
  es_opt	options;
  gi_flags	flags;

  get_capa(&options, &flags);

  /* The model */
  how.mode = 0;
  if(options.model & OPT_SQUIRT)
    how.mode |= ESQUIRT_SQUIRT;
  if(options.model & OPT_BEACON)
    how.mode |= ESQUIRT_BEACON;

  /* The transport */
  how.transport = 0;
  if(options.transport & OPT_ULTRA)
    how.transport |= ESQUIRT_ULTRA;
  if(options.transport & OPT_IRDA)
    how.transport |= ESQUIRT_IRDA;

  /* The protocol */
  how.protocol = 0;
  if(options.protocol & OPT_RAW)
    how.protocol |= ESQUIRT_RAW;
  if(options.protocol & OPT_OBEX)
    how.protocol |= ESQUIRT_OBEX;
  if(options.protocol & OPT_OBEX_GET)
    how.protocol |= ESQUIRT_OBEX_GET;
}
