/*
 *			    Squirt RUI processing
 *			    ---------------------
 *
 *		Jean II '00 & '01
 *
 * Process URI strings as found in HTML pages, and other bits
 * of magic to support RUI (Reflected UI) functionality...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/squirt_rui.c,v 2.12 2001/05/24 22:14:08 jeant Exp $
 * $Log: squirt_rui.c,v $
 * Revision 2.12  2001/05/24 22:14:08  jeant
 * Implement "uobex:/" prefix, fix documentation
 *
 * Revision 2.11  2001/05/24 18:41:36  jeant
 * Last field may not be '&' terminated in rui_extract_field()
 *
 * Revision 2.10  2001/05/24 00:15:14  jeant
 * Oups ! Unparse properly '+' in rui_extract_field()
 *
 * Revision 2.9  2001/05/24 00:01:02  jeant
 * Add rui_extract_field() for Venky
 *
 * Revision 2.8  2001/05/22 23:35:46  jeant
 * Fix for Bill : Unescape XML in server
 *
 * Revision 2.7  2001/05/22 20:09:05  jeant
 * Minor fixes (including receive NULL name)
 *
 * Revision 2.6  2001/05/21 21:34:23  jeant
 * Finishing up RUI changes (obex_type, name) and finish update compat.
 *
 * Revision 2.5  2001/05/18 23:50:29  jeant
 * Jeff changed its mind on RUI, I changed my code. It works Linux<->Linux, but still more testing needed...
 *
 * Revision 2.4  2001/05/18 02:43:57  jeant
 * First pass on latest GET/FORM encoding and overflows. Still need more testing and a few questions answered...
 *
 * Revision 2.3  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.2  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include "squirt_rui.h"
#include "squirt_xml.h"

/**************************** CONSTANTS ****************************/

/*
 * The RFC 2396 defines which character must be escaped when a URI
 * when it is embedded in HTML (2.4.3). Those are the characters we
 * need to unescape to get the proper URI.
 */
const char *	uri_control  = "\r\n";		/* All control characters */
const char *	uri_space    = " ";		/* Char 0x20 */
const char *	uri_delims   = "<>#%\"";	/* HTML delimiters */
const char *	uri_unwise   = "{}|\\^[]`";	/* Potentially dangerous */
/* Let's sum up */
const char *	uri_escape   = "\r\n <>#%\"{}|\\^[]`";

/*
 * The RFC 2396 also defines what character are reserved in URI (2.2).
 * Those characters may appear in the URI unescaped (as separating
 * the various components of the URI) or escaped (in one of the
 * component of the URI). We should *never* unescape those characters
 * to avoid messing up the original URI
 */
const char *	uri_reserved = ";/?:@&=+$,";

/*************************** URI ESCAPING ***************************/
/*
 * Subroutines to escape and unescape URI
 */

/* ---------------------------------------------------------------- */
/*
 * Escape a URI
 *
 * The source is a NULL terminated string.
 * Destination buffer plus its size
 */
char *
escape_uri(char *	dest,
	   int		dlen,
	   char *	src)
{
  char *	s = src;
  char *	d = dest;
  char *	e = dest + dlen;

  /* Look every character of the string */
  while((*s != '\0') && (d < e))
    {
      /* Is it in the list of char to escape ??? */
      if(strchr(uri_escape, *s))
	{
	  /* Escape */
	  sprintf(d, "%%%02X", *s);
	  d += 3;
	}
      else
	{
	  /* Just copy */
	  *d = *s;
	  d++;
	}
      s++;
    }

  /* Check errors */
  if(*s == '\0')
    return(dest);
  else
    return(NULL);
}

/* ---------------------------------------------------------------- */
/*
 * Un-Escape a URI
 *
 * The source is a NULL terminated string.
 * Destination buffer is at least the size of source (URI will shrink)
 * This function still works is src and dest are the same ;-)
 */
char *
unescape_uri(char *	dest,
	     char *	src)
{
  char *	s = src;
  char *	d = dest;
  char *	p;
  int		len;
  int		u;

  /* Look-up the next '%' sequence, stop when no more */
  while((p = strchr(s, '%')) != NULL)
    {
      /* Copy block of unescaped chars before the '%' */
      len = p - s;
      memcpy(d, s, len);
      d += len;
      s += len;		/* Identical to 's = p' */

      /* Deal with the escaped sequence */
      if(sscanf(s + 1, "%2X", &u) != 1)
	{
	  fprintf(stderr, "Found garbage in URI : [%c%c%c]\n",
		  *s, *(s+1), *(s+2));
	  return(NULL);	/* Oups ! Garbage... */
	}

      /* Check if it's a reserved char */
      if(strchr(uri_reserved, u))
	{
	  /* This is a reserved char, we should *not* process it */
	  memcpy(d, s, 3);
	  d += 3;
	  s += 3;
	}
      else
	{
	  /* Not reserved, un-escape it */
	  *d = u;
	  d++;
	  s+=3;
	}
    }

  /* Copy remaining of the string */
  strcpy(d, s);
  return(dest);
}

/********************* UUGLI TARGET PROCESSING *********************/
/*
 * Convert UUGLI target from human readable to binary form...
 */

/* ---------------------------------------------------------------- */
/*
 * Un-Escape a UUGLI target
 *
 * The source is a NULL terminated string.
 * Destination buffer is at least the size of source (target will shrink)
 * This function still works is src and dest are the same.
 * Note : we stop on '/' so that we can directly accept a RUI string in
 * input...
 */
int
unescape_uugli(char *	dest,
	       char *	src)
{
  char *	s = src;
  char *	d = dest;
  int		u;

  /* Look every character of the string */
  while((*s != '\0') && (*s != '/') && (*s != '<'))
    {
      /* These are natural separators of the digits, just ignore */
      if((*s == '-') || (*s == ':'))
	{
	  s++;
	  continue;
	}

      /* Grab two digits if we can */
      if(sscanf(s, "%2X", &u) != 1)
	{
	  fprintf(stderr, "Found garbage in UUGLI : [%c%c]\n",
		  *s, *(s+1));
	  return(0);	/* Oups ! Garbage... */
	}

      /* Put it in destination */
      *d = u;
      d++;
      s+=2;
    }

  /* We are cool */
  return(d - dest);
}

/*********************** HTML STRING PARSING ***********************/
/*
 * Function to parse various bits of HTML
 */

/* ---------------------------------------------------------------- */
/*
 * Look into a HREF HTML fragment containing an embedded e-squirt command
 * and extract the various pieces corresponding to this command in
 * the form expected by the library.
 *
 * Basically, you cut'n'paste the relevant part of the HTML page into
 * this function and pass the output to send_squirt()...
 *
 * We can parse these formats :
 *	embedded links (HTTP emulation, remote browsing) :
 *		obex:/index.html
 *		obex:/http://www.hp.com/index.html
 *		obex:0F1E2D3C/index.html
 *		obex:/<reference url="index.html"/>
 *		obex:0F1E2D3C/<reference url="index.html"/>
 *	embedded commands (e-Squirt emulation + backward compat) :
 *		squirt:index.html
 *		squirt:<reference url="index.html"/>
 *	embedded ultra commands (e-Squirt emulation + backward compat) :
 *		uobex:/index.html
 *		uobex:/<reference url="index.html"/>
 * We can also parse the same formats without the "obex:" or "squirt:"
 * prefix if the caller set the proper options...
 */
int
rui_parse_href(rui_data *	html,		/* The RUI string */
	       rui_params *	params,		/* Processing options */
	       es_squirt *	content,	/* Tx Payload */
	       es_opt *		options,	/* How to process */
	       es_trans *	extra,		/* Transport data */
	       gi_obget *	reply,		/* Rx Obex headers */
	       gi_flags *	flags)		/* Rx flags */
{
  char *	pointer = html->action;

  /* Note : it is the responsability of the user to clear up the
   * various structures before calling us */

  /* Check if we have an "obex:" prefix */
  if(!strncmp(html->action, "obex:", 5))
    {
      /* Skip it */
      pointer = html->action + 5;
      params->prefix = OPT_RUI_PREFIX_OBEX;
    }
  /* Check if we have an "squirt:" prefix */
  if(!strncmp(html->action, "squirt:", 7))
    {
      /* Skip it */
      pointer = html->action + 7;
      params->prefix = OPT_RUI_PREFIX_SQUIRT;
    }
  /* Check if we have an "uobex:" prefix */
  if(!strncmp(html->action, "uobex:", 6))
    {
      /* Skip it */
      pointer = html->action + 6;
      params->prefix = OPT_RUI_PREFIX_UOBEX;
    }
   /* Check that we know the prefix */
  if(params->prefix == OPT_UNDEFINED)
    {
      fprintf(stderr, "Prefix unknown...\n");
      return(-1);
    }

  /* "obex:" is followed by the target, "squirt:" is currently not */
  if(params->prefix != OPT_RUI_PREFIX_SQUIRT)
    {
      /* Extract the target */
      extra->targetlen = unescape_uugli(extra->target, pointer);
      if(extra->targetlen > 0)
	flags->obex.target = OPT_OBEX_TARGET_USER;

      /* Skip the target + ":" */
      pointer = strchr(pointer, '/');
      pointer++;
    }

  /* Use content->xml as a scratch space to unescape the URI */
  if(strlen(pointer) >= ES_MAX_XML)
    return(-2);
  /* Unescape the URI (doesn't touch already unescaped URI) */
  if(!unescape_uri(content->xml, pointer))
    return(-3);

  /* Set the proper transport flags */
  options->transport = OPT_IRDA;
  options->protocol = OPT_OBEX_GET;
  flags->transport = OPT_IRDA;
  flags->protocol = OPT_OBEX_GET;

  /* Now, check the type of string we have */
  switch(params->prefix)
    {
    case OPT_RUI_PREFIX_OBEX:
      /* Enbedded Link */

      /* BIG Warning !!!!
       * Because URL might overflow the NAME header (Obex *must* fit each
       * individual Obex header in one Obex packet, apart from Body),
       * we put the NAME field in the *body* of the packet... */
      reply->data = content->xml;
      reply->datalen = strlen(content->xml);
      reply->cleanup = OPT_DATA_NO_FREE;
      /* No squirt data, no name */
      flags->squirt_get = OPT_NO_SQUIRT_GET;
      options->do_xml = OPT_NO_XML;     /* We already have a XML string */
      flags->obex.name = OPT_OBEX_NO_NAME;
      extra->filename[0] = '\0';
      /* Set the type */
      flags->obex.type = OPT_OBEX_TYPE_NONE;	/* Indicate GET overflow */
      break;

    case OPT_RUI_PREFIX_SQUIRT:
      /* Command/reference to push out */

      /* The reference is already in content->xml, just set the flags */
      options->do_xml = OPT_NO_XML;	/* We already have a XML string */
      options->model = OPT_SQUIRT;
      /* Avoid beeing mistaken as an overflow
       * (note : setting undefined won't work because we don't do XML)
       * Also, we could skip the test and always say REFX */
      if(content->xml[0] == '<')
	options->obex.type = OPT_OBEX_TYPE_REFX;
      else
	options->obex.type = OPT_OBEX_TYPE_REFS;

      /* Enable the squirt in the GET (redundant, already the default) */
      flags->squirt_get = OPT_DO_SQUIRT_GET;
      break;

    case OPT_RUI_PREFIX_UOBEX:
      /* Ultra Command/reference to push out */

      /* The reference is already in content->xml, just set the flags */
      options->do_xml = OPT_NO_XML;	/* We already have a XML string */
      options->model = OPT_SQUIRT;
      /* Optional : set the type */
      if(content->xml[0] == '<')
	options->obex.type = OPT_OBEX_TYPE_REFX;
      else
	options->obex.type = OPT_OBEX_TYPE_REFS;
      /* For now, we totally ignore the target */

      /* Override transport to do Ultra-Obex */
      options->transport = OPT_ULTRA;
      options->protocol = OPT_OBEX;

      /* Enable the squirt in the GET (redundant, already the default) */
      flags->squirt_get = OPT_DO_SQUIRT_GET;
      break;

    default:
      fprintf(stderr, "Unknown prefix type\n");
      return(-1);
    }

  /* Tell the caller that we did it ! */
  return(flags->squirt_get);
}

/* ---------------------------------------------------------------- */
/*
 * Look into a ACTION HTML FORM fragment containing an embedded e-squirt
 * command and extract the various pieces corresponding to this command in
 * the form expected by the library.
 *
 * Basically, you cut'n'paste the relevant part of the HTML page into
 * this function and pass the output to send_squirt()...
 *
 * We can parse these formats :
 *	[Basically GET/POST forms with "obex:" prefix]
 *
 * We can also parse the same formats without the "obex:" prefix if the
 * caller set the proper options...
 */
int
rui_parse_action(rui_data *	html,		/* The RUI string */
		 rui_params *	params,		/* Processing options */
		 es_squirt *	content,	/* Tx Payload */
		 es_opt *	options,	/* How to process */
		 es_trans *	extra,		/* Transport data */
		 gi_obget *	reply,		/* Rx Obex headers */
		 gi_flags *	flags)		/* Rx flags */
{
  char *	pointer = NULL;
  int		squirt_get;
  int		name_len;

  /* Note : it is the responsability of the user to clear up the
   * various structures before calling us */

  /* Do 99% of the job... */
  squirt_get = rui_parse_href(html, params,
			      content, options, extra, reply, flags);
  if(squirt_get < 0)
    return(squirt_get);	/* Not a GET */
  /* At this point, we have the action in the name header and the proper
   * flags all over the place. We just need to deal with user content */

  switch(params->browser)
    {
    case OPT_RUI_BROWSER_NONE:
      /* No browser specified, action and input are neatly separated */

      /* Process GET Forms */
      if(params->form == OPT_RUI_FORM_GET)
	{
	  /* GET forms : the user input has to be added to the name,
	   * which is in fact in the body ;-)
	   * Fudge a bit to get the right behavior... */
	  name_len = strlen(content->xml);
	  if((name_len + strlen(html->input)) < ES_MAX_XML)
	    {
	      content->xml[name_len] = '?';
	      strcpy(content->xml + name_len + 1, html->input);
	      reply->datalen = strlen(reply->data);
	    }
	  else
	    {
	      fprintf(stderr, "Input too long to fit in body...\n");
	      return(-5);
	    }
	}
      /* Process POST Forms with overflow */
      if(params->form == OPT_RUI_FORM_POST)
	{
	  int	tot_len;
	  /* POST forms : the action and the user input has to go in the
	   * body/payload of the Obex GET request.
	   * It look exactly the same as above, apart from the type.
	   * Just for fun, we do it differently with a malloc. The
	   * reason is that post forms might be big */
	  tot_len = strlen(content->xml) + 1 + strlen(html->input);
	  reply->data = (char *) malloc(tot_len + 1);
	  if(reply->data != NULL)
	    {
	      /* Set the content */
	      strcpy(reply->data, content->xml);
	      strcat(reply->data, "?");
	      strcat(reply->data, html->input);
	      reply->datalen = strlen(reply->data);
	      reply->cleanup = OPT_DATA_DO_FREE;
	      /* Set the type */
	      if((html->encoding == NULL) || (html->encoding[0] == '\0') ||
		 (strlen(html->encoding) >= ES_MAX_TYPE))
		flags->obex.type = OPT_OBEX_TYPE_UFORM;
	      else
		{
		  strcpy(extra->type, html->encoding);
		  flags->obex.type = OPT_OBEX_TYPE_USER;
		}
	      /* Cleanup XML to be really nice */
	      content->xml[0] = '\0';
	    }
	  else
	    {
	      fprintf(stderr, "Malloc failed...\n");
	      return(-5);
	    }
	}
      break;
    case OPT_RUI_BROWSER_EUDORA:
      /* This is just for Palm compliance, the user doesn't know the type
       * of form and we have to parse the string... */

      /* Check if we have a GET form (approximative) */
      if(strchr(content->xml, '?'))
	{
	  /* Nothing else to do... */
	  params->form = OPT_RUI_FORM_GET;
	}
      /* Check if we have a POST form */
      if((pointer = strstr(content->xml, "\nPOST\n")) != NULL)
	{
	  reply->data = (char *) malloc(strlen(content->xml) + 1);
	  if(reply->data != NULL)
	    {
	      /* Remove the input from name */
	      *pointer = '\0';

	      /* Set the content */
	      strcpy(reply->data, content->xml);
	      strcat(reply->data, "?");
	      strcat(reply->data, pointer + 6);
	      reply->datalen = strlen(reply->data);
	      reply->cleanup = OPT_DATA_DO_FREE;
	      /* Set the type */
	      flags->obex.type = OPT_OBEX_TYPE_UFORM;
	      /* Cleanup XML to be really nice */
	      content->xml[0] = '\0';
	    }
	  else
	    {
	      fprintf(stderr, "Malloc failed...\n");
	      return(-5);
	    }

	  params->form = OPT_RUI_FORM_POST;
	}
      break;
    default:
      /* Bah... */
      break;
    }

  /* Tell the caller that we did it ! */
  return(squirt_get);
}

/* ---------------------------------------------------------------- */
/*
 * Look into what was received and extract the relevant bits
 * at the right place...
 * Note : we don't allocate anything here, we just point to it in
 * the other structures. Don't free request->data too early...
 */
int
rui_extract_obget(es_squirt *	content,	/* Tx Payload */
		  es_opt *	options,	/* How to process */
		  es_trans *	extra,		/* Transport data */
		  gi_obget *	reply,		/* Rx Obex headers */
		  gi_flags *	flags,		/* Rx flags */
		  rui_data *	html,		/* The RUI string */
		  rui_params *	params)		/* Processing options */
{
  char *	pointer;
  int		ret = 0;

  /* Do always generic processing */
  params->browser = OPT_RUI_BROWSER_NONE;
  params->prefix = OPT_UNDEFINED;
  params->form = OPT_UNDEFINED;
  /* Quick init */
  html->action = NULL;
  html->input = NULL;
  html->encoding = NULL;

  /* First, check if we have something in the name */
  if((flags->obex.name != OPT_OBEX_HAS_NAME) ||
     (extra->filename[0] == '\0'))
    {
      /* Sanity check */
      if((reply->data == NULL) || (reply->datalen < 0))
	return(-3);

      /* Second, check for references */
      if((reply->data[0] == '<') || (!strncasecmp(reply->data, "%3c", 3)))
	{
	  /* Check if user wants it */
	  if(content != NULL)
	    {
	      /* Check that it fits */
	      if(reply->datalen >= ES_MAX_XML)
		return(-2);

	      /* Check if escaped or not */
	      if(reply->data[0] == '<')
		/* Put XML at the right place */
		memcpy(content->xml, reply->data, reply->datalen);
	      else
		/* Unescape the XML */
		if(!unescape_uri(content->xml, reply->data))
		  return(-3);

	      /* Decapsulate XML */
	      ret = parse_xml(content, options);
	    }
	  params->form = OPT_RUI_FORM_SQUIRT;
	}
      else
	{
	  /* No name : everything in the body to avoid overflows */

	  /* Action is always at the start of the body */
	  html->action = reply->data;

	  /* Check if it looks like some form data */
	  pointer = strchr(html->action, '?');
	  if(pointer)
	    html->input = pointer + 1;

	  /* Maybe there is a type */
	  if((flags->obex.type != OPT_UNDEFINED) &&
	     (flags->obex.type != OPT_OBEX_TYPE_NONE))
	    html->encoding = extra->type;

	  /* Check the type */
	  switch(flags->obex.type)
	    {
	    case OPT_OBEX_TYPE_NONE:
	    case OPT_OBEX_TYPE_GETOV:
	      /* That's a URL - either a simple URL or a GET form */
	      if(pointer)
		params->form = OPT_RUI_FORM_GET;
	      else
		params->form = OPT_RUI_FORM_NONE;
	      break;

	    default :
	      /* That's a POST - Split action from input */
	      if(pointer)
		*pointer = '\0';
	      params->form = OPT_RUI_FORM_POST;
	      break;

	    }	/* switch() */
	}	/* if(reference) */
    }		/* if(name) */
  else
    {
      /* Simple here : action in name, input in body */
      html->action = extra->filename;
      /* Those may or may not exist */
      if(reply->datalen >= 0)
	html->input = reply->data;
      if((flags->obex.type != OPT_UNDEFINED) &&
	 (flags->obex.type != OPT_OBEX_TYPE_NONE))
	html->encoding = extra->type;
	
      /* Check the type */
      switch(flags->obex.type)
	{
	case OPT_OBEX_TYPE_UFORM:
	case OPT_OBEX_TYPE_FORM:
	  params->form = OPT_RUI_FORM_POST;
	  break;

	case OPT_OBEX_TYPE_GETOV:
	case OPT_OBEX_TYPE_POSTOV:
	  /* Should never get thoses !!! */
	  fprintf(stderr, "Invalid Obex type -> no overflow...\n");
	  params->form = OPT_UNDEFINED;
	  break;

	default:
	  /* Check how the data looks like... */
	  if(html->input)
	    /* Hu ? Probably a mangled GET form */
	    params->form = OPT_RUI_FORM_GET;
	  else
	    {
	      /* Check if it's a GET form or a simple URL */
	      pointer = strchr(html->action, '?');
	      if(pointer)
		{
		  html->input = pointer + 1;
		  params->form = OPT_RUI_FORM_GET;
		}
	      else
		params->form = OPT_RUI_FORM_NONE;
	    }
	  break;
	}
    }
  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Extract a field out of a form string
 *
 * The source is a NULL terminated string.
 * Destination buffer is at least the size of source (field will shrink)
 * This function still works is src and dest are the same ;-)
 */
char *
rui_extract_field(char *	input,		/* Input string */
		  rui_field *	field)		/* Field content */
{
  char *	s;
  char *	d;
  char *	p;
  char *	ret = NULL;
  int		len;
  int		u;

  field->label = NULL;
  field->content = NULL;

  /* Look for the end of input */
  p = strchr(input, '&');
  if(p != NULL)
    {
      /* Split input from rest of string */
      *p = '\0';
      ret = p+1;
      if(*ret == '\0')
	ret = NULL;
    }
  else
    ret = NULL;

  /* Look up for an '=' */
  p = strchr(input, '=');
  if(p == NULL)
    return(ret);

  /* Split name from input */
  *p = '\0';
  field->label = input;
  field->content = p+1;

  /* First, deal with the conversion '+' -> ' ' */
  s = field->content;
  while((p = strchr(s, '+')) != NULL)
    {
      *p = ' ';
      s = p;
    }

  /* Now, do the full unescaping of %XX characters */
  s = field->content;
  d = field->content;

  /* Look-up the next '%' sequence, stop when no more */
  while((p = strchr(s, '%')) != NULL)
    {
      /* Copy block of unescaped chars before the '%' */
      len = p - s;
      memcpy(d, s, len);
      d += len;
      s += len;		/* Identical to 's = p' */

      /* Deal with the escaped sequence */
      if(sscanf(s + 1, "%2X", &u) != 1)
	{
	  fprintf(stderr, "Found garbage in URI : [%c%c%c]\n",
		  *s, *(s+1), *(s+2));
	  return(NULL);	/* Oups ! Garbage... */
	}

      /* No character is reserved anymore, un-escape all */
      *d = u;
      d++;
      s+=3;
    }

  /* Copy remaining of the string */
  strcpy(d, s);

  return(ret);
}


/*
 * To think about :
 * A function that would remap in a HTML page all occurences
 * of "obex:" and "squirt:" to a local web server.
 * Example :
 *	obex_remap(html, "http://localhost:4242")
 *		obex:XXX => http://localhost:4242/XXX
 *		squirt:XXX => http://localhost:4243/XXX
 * Note : can't remap both obex: and squirt: to same port because
 * of syntax differences (for example initial '/')
 * It seem easier to do a perl script for that...
 */
