/*
 *			    Squirt Socket common
 *			    --------------------
 *
 *		Jean II '00
 *
 * Handle all the common code for the various e-squirt applications.
 * Squirt over IrDA sockets
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/squirt_socket.c,v 2.6 2001/05/09 21:30:39 jeant Exp $
 * $Log: squirt_socket.c,v $
 * Revision 2.6  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.5  2001/04/19 18:56:25  jeant
 * Temporary release, not fully functional
 * Reworked native API
 * Got type/accept/target Obex headers working properly
 * Updated tests
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include "squirt_socket.h"

/************************ CONSTANTS & MACROS ************************/

/*
 * This is the Ir service name (in IrIAS) that we are using for
 * e-Squirt directly on top of Ir-Socket.
 */
#define BEACON_SERVICE_NAME	"hp:beacon"
#define SQUIRT_SERVICE_NAME	"hp:esquirt"

/* Memory allocation for discovery */
#define DISC_MAX_DEVICES 10
#define DISC_BUF_LEN	sizeof(struct irda_device_list) + \
			sizeof(struct irda_device_info) * DISC_MAX_DEVICES

/************************* SOCKET TRANSPORT *************************/
/*
 * This is where we actually send and receive beacons and e-Squirt
 * over the IrDA-Ultra transport, without using Obex.
 * Under Linux, we simply use the excellent socket support ;-)
 */

/* ---------------------------------------------------------------- */
/*
 * Discover the server address...
 */
int
discover(int		fd,
	 char *		nickname,
	 char *		servicename,
	 int		filter)
{
  struct irda_device_list *	list;		/* List of device */
  unsigned char		buf[DISC_BUF_LEN];	/* Actual memory allocation */
  struct irda_ias_set	ias_query;
  int			daddr = 0x0;
  int			len;
  int			hints;
  int			i;
  int			err;

  /* Set the filter used for performing discovery */
  hints = HINT_COMPUTER | HINT_PDA;
  /* Note : any device may implement squirt over socket (like printers),
   * so let's not restrict ourselves...
   * Jean II */
#if 0
  if (setsockopt(fd, SOL_IRLMP, IRLMP_HINT_MASK_SET, &hints, sizeof(hints))) 
    return(-1);
#endif

  /* Set the list to point to the correct place */
  list = (struct irda_device_list *) buf;
  len = DISC_BUF_LEN;

  /* Ask for the discovery log */
  if(getsockopt(fd, SOL_IRLMP, IRLMP_ENUMDEVICES, buf, &len))
    {
      /* Discovery log empty (normal case) or error */
      printf("No device in sight !\n");
      return(-1);
    }
  /* Is there any addresses ? */
  if(list->len <= 0)
      /* Discovery log empty (exceptional case) */
      return(-1);

  /* Check based on nickname */
  if(nickname != NULL)
    {
      /* Go through the list */
      for (i=0; i < list->len; i++)
	{
	  /* Try to match the nickname */
	  if(!strcmp(list->dev[i].info, nickname))
	    /* Matched ! -> Return the address */
	    return(daddr);
	}
      /* No match */
      return(-1);
    }

  /* Any filtering required ??? */
  if(!filter)
    {
      /* Let's do it *really* simple */
      return(list->dev[0].daddr);
    }

  /* Go through the list and filter */
  for (i=0; i < list->len; i++)
    {
      /* Note : the following bit of code may give you troubles, comment out
       * if needed - Jean II */
      /* Ask if the requested service exist on this device */
      len = sizeof(ias_query);
      ias_query.daddr = list->dev[i].daddr;
      strcpy(ias_query.irda_class_name, servicename);
      strcpy(ias_query.irda_attrib_name, "IrDA:TinyTP:LsapSel");
      err = getsockopt(fd, SOL_IRLMP, IRLMP_IAS_QUERY, &ias_query, &len);

      /* Check for errors */
      if(err == 0)
	{
	  /* No error : Service exist on this device */
	  return(list->dev[i].daddr);
	}
      else
	{
	  if(errno != EADDRNOTAVAIL)
	    {
	      /* Various error codes : Bug in the IrDA stack or link cut */
	      printf("Bug in IrDA stack :-(\n");
	      return(-1);
	    }
	  else
	    /* Not available : Service not found on device */
	    ;	/* Just go on */
	}
    }
  /* End of discovery log... */
  printf("Could not find any suitable device...\n");
  return(-1);
}

/* ---------------------------------------------------------------- */
/*
 * Send a squirt to the server...
 */
int
send_socket(es_squirt *	content,
	    es_opt *	options,
	    es_trans *	extra,
	    int		opt_squirt)
{
  int			sock;		/* The socket itself */
  struct sockaddr_irda	peer;		/* Client connected to us */
  char *		nickname;	/* Destination name */
  int			daddr;		/* Destination address */
  int			ret = 0;

  /* Check if beacon or e-Squirt */
  if(opt_squirt == OPT_BEACON)
    strncpy(peer.sir_name, BEACON_SERVICE_NAME, 25);
  else
    strncpy(peer.sir_name, SQUIRT_SERVICE_NAME, 25);

  /* Check if nickname */
  if((options->nickname == OPT_DO_NICKNAME) && (extra != NULL))
    nickname = extra->nickname;
  else
    nickname = NULL;

  /* Create socket - Stream type */
  sock = socket(AF_IRDA, SOCK_STREAM, 0);
  if(sock < 0)
    {
      perror("Creating IrDA socket (no IrDA stack ?) ");
      return(-1);
    }

  /* Discover an address */
  daddr = discover(sock, nickname, peer.sir_name,
		   options->filter == OPT_DO_FILTER);
  if(daddr == -1)
    return(-1);

  /* Try to connect */
  peer.sir_family = AF_IRDA;
  peer.sir_lsap_sel = LSAP_ANY;
  peer.sir_addr = daddr;
  if(connect(sock, (struct sockaddr*) &peer, sizeof(struct sockaddr_irda)))
    {
      perror("Connect IrDA socket");
      return -1;
    }

  /* Send the data excluding a trailing '\0' */
  ret = write(sock, content->xml, strlen(content->xml));
  if(ret != (strlen(content->xml)))
     {
       perror("Write IrDA socket");
       return(-1);
     }

  /* The end */
  close(sock);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Handle the reception of some data on the IrDA socket
 * Just create a data socket via accept
 */
int
socket_accept(es_listen *	desc,
	      es_rx_data *	rx_data)
{
  struct sockaddr_irda	peer;		/* Client connected to us */
  int			addrlen;
  int			conn_sock;	/* The data socket */

  /* Duplicate the socket for receiving data */
  addrlen = sizeof(struct sockaddr_irda);
  conn_sock = accept(desc->sock, (struct sockaddr *) &peer, &addrlen);
  if(conn_sock <= 0)
    {
      perror("accept");
      return(-1);
    }

  /* Save the data socket for later use */
  desc->rx_sock = conn_sock;

  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Post processing handler for IrDA socket...
 * Receive and parse the data appropriately...
 */
int
socket_recv(es_listen *		desc,
	    es_rx_data *	rx_data)
{
  es_squirt *		content = rx_data->content;
  es_opt *		options = rx_data->options;
  int			len = 0;	/* Length of data received */
  int			ret;

  /* Read all packets that the client want to send */
  do
    {
      /* Read one packet */
      ret = recv(desc->rx_sock, content->xml, sizeof(content->xml), 0);
      if(ret <= 0)
	break;

      /* Save length of last non-zero packet */
      len = ret;
      if(len == sizeof(content->xml))
	{
	  fprintf(stderr, "Buffer to small - oups !\n");
	  return(-1);
	}
    }
  while(ret > 0);

  /* Check that we got something */
  if(len <= 0)
    return(-2);
  content->xml[len] = '\0';	/* Properly terminate the string */

  /* Set some flags */
  options->transport = OPT_IRDA;
  options->protocol = OPT_RAW;
  options->model = desc->opt;

  return(0);	/* Callback not called */
}

/* ---------------------------------------------------------------- */
/*
 * Destroy the ultra receiver. Close its socket.
 */
int
socket_end_recv(es_listen *	desc)
{
  /* Close the socket */
  if(desc->rx_sock >= 0)
    close(desc->rx_sock);
  desc->rx_sock = -1;
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Destroy a ultra listener. Close the socket and free pieces.
 */
int
socket_destroy(es_listen *	desc)
{
  /* Close the sockets */
  if(desc->sock >= 0)
    close(desc->sock);
  if(desc->rx_sock >= 0)
    close(desc->rx_sock);
  /* Free pieces */
  desc->next = (es_listen *) NULL;
  free(desc);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Add a receiver description for IrDA socket to the list
 */
es_listen *
add_socket(es_listen *	list,
	   es_opt *	options,
	   int		opt_squirt)
{
  es_listen *		desc;
  struct sockaddr_irda	self;		/* Socket address */
  int			sock;		/* Socket itself */

  /* Create socket - Stream type */
  sock = socket(AF_IRDA, SOCK_STREAM, 0);
  if(sock < 0)
    {
      perror("Creating IrDA socket (no IrDA stack ?) ");
      return(NULL);
    }

  /* Check if beacon or e-Squirt */
  if(opt_squirt == OPT_BEACON)
    strncpy(self.sir_name, BEACON_SERVICE_NAME, 25);
  else
    strncpy(self.sir_name, SQUIRT_SERVICE_NAME, 25);

  /* Bind to the specific service we want to receive */
  self.sir_family = AF_IRDA;
  self.sir_lsap_sel = LSAP_ANY;
  if(bind(sock, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Bind IrDA socket");
      return(NULL);
    }

  /* Set the socket in listen mode (indicate that we want packets) */
  if(listen(sock, 8))
    {
      perror("Listen IrDA socket");
      return(NULL);
    }

  /* Allocate some memory for the receiver description */
  desc = (es_listen *) malloc(sizeof(es_listen));
  if(desc == NULL)
    {
      fprintf(stderr, "Malloc failed\n");
      return(NULL);
    }

  /* Fill the description */
  desc->sock = sock;
  desc->rx_sock = -1;
  desc->opt = opt_squirt;
  desc->ultra = 0;
  desc->pre_handler = &socket_accept;
  desc->post_handler = &socket_recv;
  desc->end_recv = &socket_end_recv;
  desc->destroy = &socket_destroy;

  /* Link to the list and goodbye */
  desc->next = list;
  return(desc);
}
