/*
 *			    Squirt Ultra common
 *			    -------------------
 *
 *		Jean II '00
 *
 * Handle all the common code for the various e-squirt applications.
 * Handle Ultra transport and variation (Obex over Ultra)
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/squirt_ultra.c,v 2.7 2001/05/21 21:34:23 jeant Exp $
 * $Log: squirt_ultra.c,v $
 * Revision 2.7  2001/05/21 21:34:23  jeant
 * Finishing up RUI changes (obex_type, name) and finish update compat.
 *
 * Revision 2.6  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.5  2001/04/19 18:56:25  jeant
 * Temporary release, not fully functional
 * Reworked native API
 * Got type/accept/target Obex headers working properly
 * Updated tests
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include "squirt_ultra.h"

/****************************** TYPES ******************************/
/*
 * Various structures and type we use internally...
 */

/*
 * Content of a Ultra Obex command, broken down
 */
typedef struct es_ultra
{
  char *	payload;	/* XML string, see above */
  unsigned int	seq;		/* Sequence number */
  char *	usertype;	/* User defined type, if any */
} es_ultra;

/**************************** VARIABLES ****************************/
/* Maximum of squirt sequence number we keep in history */
#define MAX_DUP_IDS	3
/* Squirt sequence number history */
unsigned int	dup_id[MAX_DUP_IDS];
/* Current index in the list */
int		dup_idx = 0;
int		dup_num = 0;

/************************* RANDOM UTILITIES *************************/
/*
 * Various utilities that don't fall in a specific category...
 */

/* ---------------------------------------------------------------- */
/*
 * Map the field of various structures to point to each other
 */
static inline void
map_ultra_struct_fields(es_squirt *	content,
			es_trans *	extra,
			es_ultra *	uobex)
{
  memset(uobex, 0, sizeof(es_ultra));

  uobex->payload = content->xml;
  if(extra != NULL)
    uobex->usertype = extra->type;
}

/*********************** ULTRA OBEX HANDLING ***********************/
/*
 * The function here handle crude (but effective) Obex generation and
 * parsing for Ultra.
 * Obex over Ultra is much more simple than the full ultra, because it
 * is connectionless (single packet) and use a limited number of
 * headers...
 */

/* ---------------------------------------------------------------- */
/*
 * Add a Ultra Obex header containing a string
 * The string is *not* NULL terminated...
 */
static inline char *
add_uobex_string(unsigned char *	buf,
		 unsigned char		id,
		 const char *		string,
		 int			len)
{
  if(len == 0)
    len = strlen(string) + 1;	/* Include a NULL */

  buf[0] = id;				/* Obex header type */
  buf[1] = (len + 3) >> 8;		/* Obex header length - high */
  buf[2] = (len + 3) & 0xFF;		/* Obex header length - low */
  memcpy(buf + 3, string, len);		/* Obex header payload */

  return(buf + 3 + len);
}

/* ---------------------------------------------------------------- */
/*
 * Add a Ultra Obex header containing a number
 */
static inline char *
add_uobex_int(unsigned char *	buf,
	      unsigned char	id,
	      int		seq)
{
  buf[0] = id;			/* Obex header type */
  buf[1] = (seq >> 24) & 0xFF;		/* Obex header payload */
  buf[2] = (seq >> 16) & 0xFF;		/* Obex header payload */
  buf[3] = (seq >> 8) & 0xFF;		/* Obex header payload */
  buf[4] = seq & 0xFF;			/* Obex header payload */

  return(buf + 5);
}


/* ---------------------------------------------------------------- */
/*
 * Create a Ultra-Obex Beacon or E-Squirt
 *
 * IMPORTANT : this is just a test sample, and doesn't represent how
 * a beacon or e-Squirt should be, because Obex is flexible.
 * For example, we might remove the type header (optional) and the headers
 * may be swapped around.
 * The string is either a plain URL or a XML description.
 */
int
create_uobex(unsigned char *	buf,
	     es_ultra *		content,
	     es_opt *		options)
{
  unsigned char *	pos;
  int			avail;
  int			length;
  int			total;
  const char *		obex_string;

  /* Keep space for Obex command */
  pos = buf + 4;
  /* Add the target header */
  obex_string = get_es_obex_target(options, &length);
  if(obex_string != NULL)
    pos = add_uobex_string(pos, 0x46, obex_string, length);
  /* Check for e-Squirt, we need a sequence number */
  if(options->model == OPT_SQUIRT)
    /* Add the sequence number */
    pos = add_uobex_int(pos, 0xCC, content->seq);
  /* Add the type header (optional) note : need '\0' */
  obex_string = get_es_obex_type(options, content->usertype);
  if(obex_string != NULL)
    pos = add_uobex_string(pos, 0x42, obex_string, 0);
  /* Truncate payload if necessary */
  length = strlen(content->payload);
  avail = buf + 381 - pos;	/* 381 = Ultra - SAR byte */
  if(length > avail)
    {
      length = avail;
      return(-1);
    }
  /* Add the end of body header */
  pos = add_uobex_string(pos, 0x49, content->payload, length);

  /* Calculate size of Obex payload */
  total = pos - buf - 1;
  /* Add SAR + the OBEX command */
  buf[0] = 0x00;		/* Ultra SAR - Segmentation and reassembly */
  buf[1] = 0x82;		/* Obex command = put+final */
  buf[2] = total >> 8;		/* Obex command length - high */
  buf[3] = total & 0xFF;	/* Obex command length - low */

  return(pos - buf);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an Obex header
 */
char *
parse_uobex_header(unsigned char *	buf,
		   unsigned char *	end,
		   char *		payload,
		   int *		plen)
{
  int			len;

  /* Get and check length */
  len = (buf[1] << 8) | (buf[2]);
  if((buf + len) > end)
    {
      fprintf(stderr, "Invalid Obex header length (%d for 0X%X)\n", len, buf[0]);
      return(NULL);
    }
  /* Extract the payload */
  buf += 3; len -= 3;
  memcpy(payload, buf, len);
  payload[len] = '\0';		/* Properly terminate the string */

  /* Done */
  *plen = len;
  return(buf + len);
}

/* ---------------------------------------------------------------- */
/*
 * Parse the full Ultra-Obex command
 */
int
parse_uobex(unsigned char *	buf,
	    int			length,
	    es_ultra *		content,
	    es_opt *		options)
{
  unsigned char		temp[383];
  unsigned char *	pos;
  unsigned char *	end;
  int			len;
  int			ret = 0;

  /* Note : init of structures is already done for us */
  options->obex.type = OPT_OBEX_TYPE_NONE;

  /* Check the initial header */
  if(buf[0] != 0x00)
    {
      fprintf(stderr, "Invalid SAR\n");
      ret = -1;
    }
  if(buf[1] != 0x82)
    {
      fprintf(stderr, "Invalid Obex command (0x%X)\n", buf[1]);
      ret = -2;
    }
  len = (buf[2] << 8) | (buf[3]);
  if(len != (length - 1))
    {
      fprintf(stderr, "Invalid Obex command length (%d instead of %d)\n",
	      len, length - 1);
      ret = -3;
    }
  if(ret < 0)
    return(ret);

  /* Skip to the first header */
  pos = buf + 4;
  end = buf + length;

  /* Get all Obex headers */
  while(pos < end)
    {
      /* Look at the header type */
      switch(pos[0])
	{
	case 0x46:		/* Target header */
	  pos = parse_uobex_header(pos, end, temp, &len);
	  if(pos == NULL)
	    return -4;
	  match_es_obex_target(temp, len, options);
	  break;
	case 0x42:		/* Type header */
	  pos = parse_uobex_header(pos, end, temp, &len);
	  if(pos == NULL)
	    return -5;
	  match_es_obex_type(temp, len, options, content->usertype);
	  break;
	case 0xCC:		/* Apps params header -> Seq num */
	  content->seq =  ((pos[1] << 24) | (pos[2] << 16) | 
			   (pos[3] <<  8) | (pos[4]));
	  pos += 5;
	  break;
	case 0x49:		/* Body header */
	  pos = parse_uobex_header(pos, end, content->payload, &len);
	  if(pos == NULL)
	    return -6;
	  break;
	default:
	  fprintf(stderr, "Unknown header 0x%02X\n", pos[0]);
	  /* Let's attempt to skip the header */
	  switch(pos[0] & 0xC0)
	    {
	    case 0x00:	/* Unicode string */
	    case 0x40:	/* Byte octet sequence */
	      pos += (buf[1] << 8) | (buf[2]);
	      break;
	    case 0x80:	/* One byte */
	      pos += 2;
	      break;
	    case 0xC0:	/* 4 bytes */
	      pos += 5;
	      break;
	    }
	  break;
	}
    }
  return 0;
}

/************************* ULTRA TRANSPORT *************************/
/*
 * This is where we actually send and receive beacons and e-Squirt
 * over the IrDA-Ultra transport.
 * Under Linux, we simply use the excellent socket support ;-)
 */

/* ---------------------------------------------------------------- */
/*
 * Send a ultra packet
 * We create and close the socket here
 */
int
send_ultra(es_squirt *	content,
	   es_opt *	options,
	   es_trans *	extra)
{
  int			sock;		/* The socket itself */
  struct sockaddr_irda	self;		/* Socket address */
  es_ultra		command;	/* The ultra packet to send */
  unsigned char		buf[383];	/* Send buffer */
  int			length;
  int			actual;

  map_ultra_struct_fields(content, extra, &command);
  if(extra != NULL)
    command.seq = extra->seq;
  else
    command.seq = 0;

  /* Check if Obex */
  if(options->protocol == OPT_OBEX)
    {
      /* Obex : create Obex command */
      length = create_uobex(buf, &command, options);
      if(length < 0)
	return(length);
    }
  else
    {
      /* Raw : Send directly the XML string */
      length = strlen(command.payload);
      if(length > 381)
	return(-1);
      memcpy(buf, command.payload, length);
    }

  /* Create socket  - Ultra type */
  sock = socket(AF_IRDA, SOCK_DGRAM, 1);
  if(sock < 0)
    {
      perror("Creating ultra socket");
      return(-1);
    }

  /* Bind to the specific Ultra PID */
  self.sir_family = AF_IRDA;
  self.sir_lsap_sel = ((options->protocol == OPT_OBEX)
		       ? UPID_OBEX : UPID_BEACON);
  if(bind(sock, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Connect ultra socket");
      return(-1);
    }

  /* Send the stuff */
  actual = send(sock, buf, length, 0);

  /* Close the socket */
  close(sock);

  /* Check */
  if(actual != length)
    {
      fprintf(stderr, "Ultra socket send failed (%d)\n", actual);
      return(-1);
    }
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Handle the reception of some data on the Ultra socket
 * Receive and parse the data appropriately...
 */
int
ultra_handler(es_listen *	desc,
	      es_rx_data *	rx_data)
{
  es_squirt *		content = rx_data->content;
  es_opt *		options = rx_data->options;
  int			ret;

  /* This read should not block */
  ret = recv(desc->sock, content->xml, sizeof(content->xml) - 1, 0);
  if(ret < 0)
    return(ret);
  content->xml[ret] = '\0';	/* Properly terminate the string */

  /* Set some flags */
  options->transport = OPT_ULTRA;
  options->protocol = OPT_RAW;
  options->model = OPT_UNDEFINED;

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Handle the reception of some data on the Ultra Obex socket
 * Receive and parse the data appropriately...
 */
int
uobex_handler(es_listen *	desc,
	      es_rx_data *	rx_data)
{
  es_squirt *		content = rx_data->content;
  es_opt *		options = rx_data->options;
  es_trans *		extra   = rx_data->extra;
  unsigned char		buf[383];	/* Receive buffer */
  es_ultra		command;	/* Obex comamnd */
  int			len;
  int			ret;

  map_ultra_struct_fields(content, extra, &command);

  /* This read should not block */
  len = recv(desc->sock, &buf, sizeof(buf) - 1, 0);
  if(len < 0)
    return(len);

  /* Set some flags */
  options->transport = OPT_ULTRA;
  options->protocol = OPT_OBEX;

  /* Process the Obex command */
  ret = parse_uobex(buf, len, &command, options);

  /* Transport data */
  if(extra != NULL)
    extra->seq = command.seq;

  /* Dup filtering facility, only for squirts */
  if((ret == 0) && (options->model == OPT_SQUIRT) &&
     (rx_data->listen_options->uobex_dups == OPT_UOS_DUPS_DEL))
    {
      int	i;

      /* Check if dup */
      for(i = 0; i < dup_num; i++)
	if(command.seq == dup_id[i])
	  {
	    fprintf(stderr, "Ultra-Obex dup found - discarding...\n");
	    return(-7);		/* Dup found !!! */
	  }
      /* Insert new sequence number in table */
      dup_id[dup_idx] = command.seq;
      /* Increment indexes (a bit tricky) */
      dup_idx++;
      if(dup_num < MAX_DUP_IDS)
	dup_num = dup_idx;
      if(dup_idx >= MAX_DUP_IDS)
	dup_idx = 0;
    }

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Post processing handler for ultra...
 * Finish parsing the data.
 */
int
ultra_post_handler(es_listen *	desc,
		   es_rx_data *	rx_data)
{
  /* All the receive and Ultra-Obex processing was done in the pre-handler,
   * and XML parsing will be done after us, so we don't have anything to do
   * here... */

  return(0);	/* Callback not called */
}

/* ---------------------------------------------------------------- */
/*
 * Destroy the ultra receiver. Close its socket.
 */
int
ultra_end_recv(es_listen *	desc)
{
  /* Close the socket */
  if(desc->rx_sock >= 0)
    close(desc->rx_sock);
  desc->rx_sock = -1;
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Destroy a ultra listener. Close the socket and free pieces.
 */
int
ultra_destroy(es_listen *	desc)
{
  /* Close the sockets */
  if(desc->sock >= 0)
    close(desc->sock);
  if(desc->rx_sock >= 0)
    close(desc->rx_sock);
  /* Free pieces */
  desc->next = (es_listen *) NULL;
  free(desc);
  return(0);
}

/* ---------------------------------------------------------------- */
/*
 * Add a receiver description for Ultra socket to the list
 */
es_listen *
add_ultra(es_listen *	list,
	  es_opt *	options,
	  int		upid)
{
  es_listen *		desc;
  struct sockaddr_irda	self;		/* Socket address */
  int			sock;		/* Socket itself */

  /* Create socket  - Ultra type */
  sock = socket(AF_IRDA, SOCK_DGRAM, 1);
  if(sock < 0)
    {
      perror("Creating ultra socket (no IrDA stack ?) ");
      return(NULL);
    }

  /* Bind to the specific Ultra PID */
  self.sir_family = AF_IRDA;
  self.sir_lsap_sel = upid;
  if(bind(sock, (struct sockaddr*) &self, sizeof(struct sockaddr_irda)))
    {
      perror("Connect ultra socket (no IrDA stack ?) ");
      return(NULL);
    }

  /* Allocate some memory for the receiver description */
  desc = (es_listen *) malloc(sizeof(es_listen));
  if(desc == NULL)
    {
      fprintf(stderr, "Malloc failed\n");
      return(NULL);
    }

  /* Fill the description */
  desc->sock = sock;
  desc->rx_sock = -1;
  desc->ultra = 1;
  if(upid == UPID_OBEX)
    desc->pre_handler = &uobex_handler;
  else
    desc->pre_handler = &ultra_handler;
  desc->post_handler = &ultra_post_handler;
  desc->end_recv = &ultra_end_recv;
  desc->destroy = &ultra_destroy;

  /* Link to the list and goodbye */
  desc->next = list;
  return(desc);
}
