/*
 *			    Squirt XML common
 *			    -----------------
 *
 *		Jean II '00
 *
 * Handle XML generating and parsing.
 * We support various kind of XML here...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/squirt_xml.c,v 2.7 2001/05/09 21:30:39 jeant Exp $
 * $Log: squirt_xml.c,v $
 * Revision 2.7  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.6  2001/04/06 21:44:26  jeant
 * Change "<squirt" into "<reference"
 *
 * Revision 2.5  2001/03/06 23:47:06  jeant
 * Version 2.4.4 (don't trust CVS) : Properly terminate XML ('\>' -> '/>'), Allow sending no-session GET, mandate OpenObex 0.9.8 and minor cleanups.
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

/***************************** INCLUDES *****************************/

#include "squirt_xml.h"

/****************************** TYPES ******************************/
/*
 * Various structures and type we use internally...
 */

/*
 * Content of a e-Squirt XML string, broken down
 */
typedef struct es_xml
{
  /* XML stuff */
  char *	label;		/* Readable label : "CoolTown" */
  char *	url;		/* URL : "http://www.cooltown.com" */
  char *	type;		/* Content mime type : "text/html" */
} es_xml;

/************************* RANDOM UTILITIES *************************/
/*
 * Various utilities that don't fall in a specific category...
 */

/* ---------------------------------------------------------------- */
/*
 * Map the field of various structures to point to each other
 */
static inline void
map_xml_struct_fields(es_squirt *	content,
		      es_xml *		xml)
{
  /* Map the es_xml structure */
  xml->label = content->label;
  xml->url   = content->url;
  xml->type  = content->type;
}

/************************** XML GENERATING **************************/
/*
 * The function here handle crude (but effective) XML generation.
 * The ultimate payload of an e-Squirt transaction is a snipset of XML
 * (in some case it may be a pure URL), but as it is pretty simple
 * and always the same, no need to get a "heavy" XLM library
 * involved
 */

/* ---------------------------------------------------------------- */
/*
 * Add a XML token to string
 */
static inline void
add_xml_token(char *	buf,
	      char *	label,
	      char *	token)
{
  int	len = strlen(buf);

  /* Just do it */
  sprintf(buf + len, " %s=\"%s\"", label, token);
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string (XLink format)
 */
int
create_xlink_xml(char *		buf,
		 es_xml *	content,
		 es_opt *	options)
{
  /* Start the XML string */
  strcpy(buf, "<a:link");
  /* Put the label, if needed */
  if(content->label[0] != '\0')
    add_xml_token(buf, "xlink:title", content->label);
  /* Put the url, if needed */
  if(content->url[0] != '\0')
    add_xml_token(buf, "xlink:href", content->url);
  /* Put the type, if needed */
  if(content->type[0] != '\0')
    add_xml_token(buf, "a:ContentType", content->type);
  /* Terminate XML */
  strcat(buf, "/>");

  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string (old format)
 */
int
create_old_xml(char *	buf,
	       es_xml *	content,
	       es_opt *	options)
{
  /* Start the XML string */
  strcpy(buf, "<squirt");
  /* Put the label, if needed */
  if(content->label[0] != '\0')
    add_xml_token(buf, "label", content->label);
  /* Put the url, if needed */
  if(content->url[0] != '\0')
    add_xml_token(buf, "url", content->url);
  /* Put the type, if needed */
  if(content->type[0] != '\0')
    add_xml_token(buf, "type", content->type);
  /* Terminate XML */
  strcat(buf, "/>");

  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string (standard format)
 */
int
create_std_xml(char *	buf,
	       es_xml *	content,
	       es_opt *	options)
{
  /* Start the XML string */
  strcpy(buf, "<reference");
  /* Put the label, if needed */
  if(content->label[0] != '\0')
    add_xml_token(buf, "label", content->label);
  /* Put the url, if needed */
  if(content->url[0] != '\0')
    add_xml_token(buf, "url", content->url);
  /* Put the type, if needed */
  if(content->type[0] != '\0')
    add_xml_token(buf, "type", content->type);
  /* Terminate XML */
  strcat(buf, "/>");

  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * Create VCard string (backward compatibility)
 */
int
create_vcard_xml(char *		buf,
		 es_xml *	content,
		 es_opt *	options)
{
  /* Do it the hard way... */
  if(content->label[0] != '\0')
    sprintf(buf, "BEGIN:VCARD\r\nN:%s\r\nURL:%s\r\nADR:%s\r\nEND:VCARD\r\n",
	    content->label, content->url, content->url);
  else
    sprintf(buf, "BEGIN:VCARD\r\nURL:%s\r\nADR:%s\r\nEND:VCARD\r\n",
	    content->url, content->url);

  return 0;
}

/* ---------------------------------------------------------------- */
/*
 * Create XML string (generic)
 * This just call one of the previous function depending on the args...
 */
int
create_xml(es_squirt *	content,
	   es_opt *	options)
{
  es_xml	xml;		/* The XML string to form */
  char *	buf;		/* The result */
  int		ret = 0;

  /* Map internal structs */
  map_xml_struct_fields(content, &xml);
  buf = content->xml;

  /* Check format desired */
  switch(options->xml)
    {
    case OPT_XML_URL:
      strcpy(buf, xml.url);
      break;
    case OPT_XML_OLD:
      ret = create_old_xml(buf, &xml, options);
      break;
    case OPT_XML_STD:
      ret = create_std_xml(buf, &xml, options);
      break;
    case OPT_XML_XLINK:
      ret = create_xlink_xml(buf, &xml, options);
      break;
    case OPT_XML_VCARD:
      ret = create_vcard_xml(buf, &xml, options);
      break;
    default:
      fprintf(stderr, "Invalid XML format\n");
      ret = -1;
    }
  return ret;
}

/*************************** XML PARSING ***************************/
/*
 * The function here handle crude (but effective) XML parsing.
 * Reverse of what we do for generating it...
 */

/* ---------------------------------------------------------------- */
/*
 * Find an XML token in the string
 * Note : there are a lot of cases where this won't work.
 * But in most cases common, this will do the job properly.
 */
char *
parse_token(char *	string,
	    char *	id,
	    char *	result,
	    int		max)
{
  char *		pos;
  char *		pos2;
  char *		pos3;

  /* Find the ID */
  pos = strstr(string, id);
  if(pos == NULL)
    return(NULL);

  /* Find the first `"' */
  pos2 = strchr(pos, '"');
  if(pos2 == NULL)
    return(NULL);
  pos2++;

  /* Find the second `"' */
  pos3 = strchr(pos2, '"');
  if(pos3 == NULL)
    return(NULL);

  /* Check buffer overflows */
  if((pos3 - pos2) > (max - 1))
    {
      fprintf(stderr, "Token %s too big (%d)...\n", id, pos3 - pos2);
      return(NULL);
    }

  /* Copy over */
  memcpy(result, pos2, pos3 - pos2);
  result[pos3 - pos2] = '\0';
  return(result);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string to get out some token we care about (XLink format)
 */
int
parse_xlink_xml(char *	string,
		es_xml *	content,
		es_opt *	options)
{
  int		ret = 0;

  /* Find the label */
  content->label[0] = '\0';
  parse_token(string, "xlink:title=", content->label, ES_MAX_LABEL);

  /* Find the url */
  content->url[0] = '\0';
  if(NULL == parse_token(string, "xlink:href=", content->url, ES_MAX_URL))
    ret = -1;

  /* Find the label */
  content->type[0] = '\0';
  parse_token(string, "a:ContentType=", content->type, ES_MAX_TYPE);

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string to get out some token we care about (standard format)
 */
int
parse_std_xml(char *	string,
	      es_xml *	content,
	      es_opt *	options)
{
  int		ret = 0;

  /* Find the label */
  content->label[0] = '\0';
  parse_token(string, "label=", content->label, ES_MAX_LABEL);

  /* Find the url */
  content->url[0] = '\0';
  if(NULL == parse_token(string, "url=", content->url, ES_MAX_URL))
    ret = -1;

  /* Find the label */
  content->type[0] = '\0';
  parse_token(string, "type=", content->type, ES_MAX_TYPE);

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Parse a VCard (gross, don't look)
 */
int
parse_vcard_xml(char *		string,
		es_xml *	content,
		es_opt *	options)
{
  int		ret = -1;
  char *		pos;
  char *		pos2;
  char *		pos3;

  /* Init */
  content->label[0] = '\0';
  content->url[0] = '\0';
  content->type[0] = '\0';

  /* Find the label */
  pos = strstr(string, "\nN:");
  if(pos != NULL)
    {
      /* Find the '\n' and the '\r' */
      pos2 = strchr(pos + 3, '\n');
      pos3 = strchr(pos + 3, '\r');
      if((pos3 != NULL) && (pos3 < pos2))
	pos2 = pos3;
      if(pos2 != NULL)
	{
	  /* Copy over */
	  memcpy(content->label, pos + 3, pos2 - pos - 3);
	  content->label[pos2 - pos - 3] = '\0';
	}
    }

  /* Find the url */
  pos = strstr(string, "\nURL:");
  if(pos != NULL)
    {
      /* Find the '\n' */
      pos2 = strchr(pos + 5, '\n');
      pos3 = strchr(pos + 3, '\r');
      if((pos3 != NULL) && (pos3 < pos2))
	pos2 = pos3;
      if(pos2 != NULL)
	{
	  /* Copy over */
	  memcpy(content->url, pos + 5, pos2 - pos - 5);
	  content->url[pos2 - pos] = '\0';
	  ret = 0;
	}
    }
  else
    {
      pos = strstr(string, "http:");
      if(pos != NULL)
	{
	  /* Find the '\n' */
	  pos2 = strchr(pos, '\n');
	  pos3 = strchr(pos + 3, '\r');
	  if((pos3 != NULL) && (pos3 < pos2))
	    pos2 = pos3;
	  if(pos2 != NULL)
	    {
	      /* Copy over */
	      memcpy(content->url, pos, pos2 - pos);
	      content->url[pos2 - pos] = '\0';
	      ret = 0;
	    }
	}
    }

  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Parse an XML string to get out some token we care about (generic)
 * We just call the correct function
 */
int
parse_xml(es_squirt *	content,
	  es_opt *	options)
{
  char *	string;		/* The input */
  es_xml	xml;		/* The XML string to parse */
  int		ret = 0;

  /* Map internal structs */
  string = content->xml;
  map_xml_struct_fields(content, &xml);

  /* Check if XML or not */
  if(string[0] != '<')
    {
      /* Check VCard */
      if(!strncmp(string, "BEGIN:VCARD", 11))
	{
	  /* VCard format */
	  ret = parse_vcard_xml(string, &xml, options);
	  options->xml = OPT_XML_VCARD;
	}
      else
	{
	  /* Plain URL */
	  if(strlen(string) <= ES_MAX_URL)
	    strcpy(xml.url, string);
	  else
	    ret = -1;
	  xml.label[0] = '\0';
	  xml.type[0] = '\0';
	  options->xml = OPT_XML_URL;
	}
   }
  else
    {
      /* Check the type of XML... */
      if(string[1] != 'a')
	{
	  /* Check the type of XML (bis)... */
	  if(string[1] == 'r')
	    options->xml = OPT_XML_STD;
	  else
	    options->xml = OPT_XML_OLD;
	  /* Standard XML format (or old one) */
	  ret = parse_std_xml(string, &xml, options);
	}
      else
	{
	  /* XLink XML format */
	  ret = parse_xlink_xml(string, &xml, options);
	  options->xml = OPT_XML_XLINK;
	}
    }
  return(ret);
}
