/*
 *			    IrDA Obex Squirt+Get
 *			    --------------------
 *
 *		Jean II '99 & '00
 *
 * Send and receive e-Squirt using Obex GET.
 * This is ugly and prone to confusion, but it's in the spec...
 *
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/tests/squirt_get.c,v 2.9 2001/05/18 23:50:29 jeant Exp $
 * $Log: squirt_get.c,v $
 * Revision 2.9  2001/05/18 23:50:29  jeant
 * Jeff changed its mind on RUI, I changed my code. It works Linux<->Linux, but still more testing needed...
 *
 * Revision 2.8  2001/05/18 02:43:57  jeant
 * First pass on latest GET/FORM encoding and overflows. Still need more testing and a few questions answered...
 *
 * Revision 2.7  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.6  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 * Revision 2.5  2001/04/19 18:56:25  jeant
 * Temporary release, not fully functional
 * Reworked native API
 * Got type/accept/target Obex headers working properly
 * Updated tests
 *
 * Revision 2.4.3.1  2001/02/16 23:57:34  jeant
 * Imported sources
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <squirt.h>

/*#define BROWSER		"lynx %s"*/
//#define BROWSER		"netscape %s &"

/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon. Non forking.
 */
int
receive_esquirt(es_squirt *	fields,		/* Rx payload */
		es_opt *	options,	/* Rx options */
		es_trans *	extra,		/* Rx transport data */
		gi_obget *	ret_hdr,	/* Not used */
		gi_flags *	ret_flags)	/* Not used */
{
  int		ret = 0;	/* Reply Header not set */

  debug_squirt(fields, options, extra);

#ifdef BROWSER
  {
    unsigned char		buf[512];	/* The command */
    sprintf(buf, BROWSER, fields->url);
    system(buf);
  }
#endif

  /* Check if we can return some data */
  if(ret_hdr != (gi_obget *) NULL)
    {
      /* Set the name */
      strcpy(extra->filename, "index.html");
      ret_flags->obex.name = OPT_OBEX_HAS_NAME;
      /* Set the types we support (only applicable for OPT_OBEX) */
      strcpy(extra->accept, "text/html, text/xml");
      ret_flags->obex.accept = OPT_OBEX_HAS_ACCEPT;

      ret = 1;	/* Reply Header set */
    }

  /* Return back to the server */
  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a Get Interface request.
 */
int
reply_obget(gi_obget *	request,
	    gi_flags *	flags,
	    es_trans *	extra,
	    void *	userdata)
{
  static char	string[] = "<HTML><HEAD><TITLE>My Interface</TITLE></HEAD>
<BODY>No much yet ;-)</BODY></HTML>";

  /* Check what we have */
  debug_obget(request, flags, extra);

  /* Note : 
   * If it is a simple get-interface, the requester may have set name and
   * type in its request, and we should take this into account.
   * If it is a squirt-get, receive_esquirt() will have already set
   * name and type in the structure */

  /* Get rid of potential form data passed to us */
  if((request->data != NULL) && (request->cleanup == OPT_DATA_DO_FREE))
    free(request->data);

  /* Set the name */
  strcpy(extra->filename, "index.html");
  flags->obex.name = OPT_OBEX_HAS_NAME;
  /* Set the type */
  strcpy(extra->type, "text/html");
  flags->obex.type = OPT_OBEX_TYPE_USER;

  /* Set the data */
  request->data = string;
  /* Set the length (not including '\0') */
  request->datalen = strlen(string);
  request->cleanup = OPT_DATA_NO_FREE;

  /* Return back to the server */
  return(1);
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  es_squirt	fields;
  es_opt	options;
  es_trans	extra;
  gi_obget	interface;
  gi_flags	flags;
  int		ret = -1;

  init_structs(&fields, &options, &extra, &interface, &flags, NULL);
  /* Set the options */
  options.xml = OPT_XML_STD;
  options.model = OPT_UNDEFINED;	/* Only for the receiver */
  options.transport = OPT_IRDA;
  options.protocol = OPT_OBEX_GET;
  options.obex.type = OPT_OBEX_TYPE_NONE;

  /* Look if there is any command line args */
  switch(argc)
    {
    case 0:
      /* Fallthrough */
    case 1:
      /* No argument : receive a beacon */
      break;
    case 2:
      /* One argument : the URL */
      strcpy(fields.url, argv[1]);
      break;
    case 3:
      /* Two arguments : label + url */
      strcpy(fields.label, argv[1]);
      strcpy(fields.url, argv[2]);
      break;
    default:
      /* Fallthrough */
    case 4:
      /* Three arguments : label + url + type */
      strcpy(fields.label, argv[1]);
      strcpy(fields.url, argv[2]);
      strcpy(fields.type, argv[3]);
      break;
    }

  if(argc > 1)
    {
      /* Send a single e-Squirt/Beacon and get associated interface */
      options.model = OPT_SQUIRT;	/* Make sure model is defined */
      ret = send_squirt(&fields, &options, &extra, &interface, &flags);

      /* Display it */
      debug_obget(&interface, &flags, &extra);

      if(ret < 0)
	printf("Sending e-Squirt failed...\n");
    }
  else
    {
      /* Prepare to handle get interface calls */
      set_obget_handler(&reply_obget, NULL);

      /* Receive multiple e-Squirt/Beacon */
      printf("Waiting for e-Squirts over Obex...\n");
      recv_squirt(&fields, &options, &extra, &receive_esquirt, NULL);
      /* Will loop forever */
    }

  return(ret);
}
