/*
 *			    IrDA Obex Squirt a HREF string
 *			    ------------------------------
 *
 *		Jean II '99, '00 & '01
 *
 * Parse a HREF string and send it Obex GET.
 * 
 * This file is released under the GPL license.
 */

/*
 * CVS blurb
 * $Header: /home/cvs/Hack/Esquirt/Linux/tests/squirt_href.c,v 2.7 2001/05/22 20:09:05 jeant Exp $
 * $Log: squirt_href.c,v $
 * Revision 2.7  2001/05/22 20:09:05  jeant
 * Minor fixes (including receive NULL name)
 *
 * Revision 2.6  2001/05/21 21:34:23  jeant
 * Finishing up RUI changes (obex_type, name) and finish update compat.
 *
 * Revision 2.5  2001/05/18 23:50:29  jeant
 * Jeff changed its mind on RUI, I changed my code. It works Linux<->Linux, but still more testing needed...
 *
 * Revision 2.4  2001/05/18 02:43:57  jeant
 * First pass on latest GET/FORM encoding and overflows. Still need more testing and a few questions answered...
 *
 * Revision 2.3  2001/05/09 21:30:39  jeant
 * Proper licensing terms (thanks to Mirjana)...
 *
 * Revision 2.2  2001/04/24 00:53:07  jeant
 * The good one : fix remaining bugs found during Linux<->Linux testing
 *
 *
 */
/*
Copyright (C) 1995-2001 Hewlett-Packard Company 

This program is free software; you can redistribute it and/or modify it under the terms of the GNU General Public License as published by the Free Software Foundation; either version 2 of the License, or (at your option) any later version. 

This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for more details. 

You should have received a copy of the GNU General Public License along with this program; if not, write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
*/

#include <unistd.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>

#include <squirt.h>

/*#define BROWSER		"lynx %s"*/
//#define BROWSER		"netscape %s &"

/***************************** RECEIVER *****************************/

/* ---------------------------------------------------------------- */
/*
 * Receive a e-Squirt of Beacon. Non forking.
 */
int
receive_esquirt(es_squirt *	fields,		/* Rx payload */
		es_opt *	options,	/* Rx options */
		es_trans *	extra,		/* Rx transport data */
		gi_obget *	ret_hdr,	/* Not used */
		gi_flags *	ret_flags)	/* Not used */
{
  int		ret = 0;	/* Reply Header not set */

  debug_squirt(fields, options, extra);

#ifdef BROWSER
  {
    unsigned char		buf[512];	/* The command */
    sprintf(buf, BROWSER, fields->url);
    system(buf);
  }
#endif

  /* Check if we can return some data */
  if(ret_hdr != (gi_obget *) NULL)
    {
      /* Set the name */
      strcpy(extra->filename, "index.html");
      ret_flags->obex.name = OPT_OBEX_HAS_NAME;
      /* Set the types we support (only applicable for OPT_OBEX) */
      strcpy(extra->accept, "text/html, text/xml");
      ret_flags->obex.accept = OPT_OBEX_HAS_ACCEPT;

      ret = 1;	/* Reply Header set */
    }

  /* Return back to the server */
  return(ret);
}

/* ---------------------------------------------------------------- */
/*
 * Receive a Get Interface request.
 */
int
reply_obget(gi_obget *	request,
	    gi_flags *	flags,
	    es_trans *	extra,
	    void *	userdata)
{
  static char	string[] = "<HTML><HEAD><TITLE>My Interface</TITLE></HEAD>
<BODY>No much yet ;-)</BODY></HTML>";

  /* Check what we have */
  debug_obget(request, flags, extra);

  /* Note : 
   * If it is a simple get-interface, the requester may have set name and
   * type in its request, and we should take this into account.
   * If it is a squirt-get, receive_esquirt() will have already set
   * name and type in the structure */

  /* Get rid of potential form data passed to us */
  if((request->data != NULL) && (request->cleanup == OPT_DATA_DO_FREE))
    free(request->data);

  /* Set the name */
  strcpy(extra->filename, "index.html");
  flags->obex.name = OPT_OBEX_HAS_NAME;
  /* Set the type */
  strcpy(extra->type, "text/html");
  flags->obex.type = OPT_OBEX_TYPE_USER;

  /* Set the data */
  request->data = string;
  /* Set the length (not including '\0') */
  request->datalen = strlen(string);
  request->cleanup = OPT_DATA_NO_FREE;

  /* Return back to the server */
  return(1);
}

/******************************* MAIN *******************************/

/* ---------------------------------------------------------------- */
/*
 * main body of the program
 */
int
main(int	argc,
     char *	argv[])
{
  es_squirt	fields;
  es_opt	options;
  es_trans	extra;
  gi_obget	interface;
  gi_flags	flags;
  rui_data	html;		/* The RUI string */
  rui_params	params;		/* Processing opts */
  int		ret = -1;

  init_structs(&fields, &options, &extra, &interface, &flags, NULL);

  /* Look if there is any command line args */
  if(argc > 1)
    {
      fprintf(stderr, " ---- Parsing command line... ----\n\n");

      /* Parse what we have */
      params.prefix = OPT_RUI_PREFIX_OBEX;
      html.action = argv[1];
      html.input = NULL;
      html.encoding = NULL;
      if(rui_parse_href(&html, &params,
			&fields, &options, &extra, &interface, &flags) < 0)
	return(-1);

      /* A few extra options... */
      options.obex.ias = OPT_OBEX_DO_STDIAS;
      options.obex.target = OPT_OBEX_TARGET_UUGLI;
      flags.obex.ias = OPT_OBEX_DO_STDIAS;

      /* Display what has been done */
      debug_squirt(&fields, &options, &extra);
      debug_obget(&interface, &flags, &extra);
      fprintf(stderr, " ---- Now, trying to send this... ----\n\n");

      /* Send a single e-Squirt/Beacon and get associated interface */
      ret = send_squirt(&fields, &options, &extra, &interface, &flags);

      /* Display it */
      if(ret < 0)
	printf("Sending e-Squirt failed...\n");
      else
	debug_obget(&interface, &flags, &extra);
    }
  else
    {
      /* Set the receive options that matter */
      options.model = OPT_UNDEFINED;	/* Only for the receiver */
      options.transport = OPT_IRDA;
      options.protocol = OPT_OBEX_GET;

      /* Prepare to handle get interface calls */
      set_obget_handler(&reply_obget, NULL);

      /* Receive multiple e-Squirt/Beacon */
      printf("Waiting for e-Squirts over Obex...\n");
      recv_squirt(&fields, &options, &extra, &receive_esquirt, NULL);
      /* Will loop forever */
    }

  return(ret);
}
